from abc import ABC, abstractmethod
from pathlib import Path

from loguru import logger

from infrasys.system import System


class AbstractReader(ABC):
    @abstractmethod
    def get_system(self) -> System:
        """Method return parsed Infrasys system

        Returns:
            System: Instance of Infrasys system
        """
        ...

    def to_json(self, json_file: Path | str):
        """Exports system to a json file

        Args:
            json_file (Path | str): Export path for the GDM model
        """
        json_file = Path(json_file)
        assert hasattr(self, "system"), "Use the read method to build the system first"
        self.system.to_json(json_file, overwrite=True)
        logger.debug(f"GDM model exported to {json_file}")



from collections import defaultdict
from io import TextIOWrapper
from pathlib import Path
from typing import Any


from altdss_schema import altdss_models
from gdm.distribution.components import (
    DistributionVoltageSource,
    DistributionComponentBase,
    DistributionTransformer,
    DistributionBranchBase,
    MatrixImpedanceSwitch,
    DistributionBus,
)
from gdm.distribution.equipment import (
    ConcentricCableEquipment,
    BareConductorEquipment,
)
from loguru import logger

from ditto.writers.abstract_writer import AbstractWriter
import ditto.writers.opendss as opendss_mapper
from ditto.enumerations import OpenDSSFileTypes


class Writer(AbstractWriter):
    files = []

    def _get_dss_string(self, model_map: Any) -> str:
        # Example model_map is instance of DistributionBusMapper
        altdss_class = getattr(altdss_models, model_map.altdss_name)
        # Example altdss_class is Bus
        altdss_object = altdss_class.model_validate(model_map.opendss_dict)
        if model_map.altdss_composition_name is not None:
            altdss_composition_class = getattr(altdss_models, model_map.altdss_composition_name)
            altdss_composition_object = altdss_composition_class(altdss_object)
            dss_string = altdss_composition_object.dumps_dss()
        else:
            dss_string = altdss_object.dumps_dss()

        return dss_string

    def prepare_folder(self, output_path):
        directory = Path(output_path)
        files_to_remove = directory.rglob("*.dss")
        for dss_file in files_to_remove:
            logger.debug(f"Deleting existing file {dss_file}")
            # dss_file.unlink() #TODO: deletion causing tets to fail @tarek

    def _get_voltage_bases(self) -> list[float]:
        voltage_bases = []
        buses: list[DistributionBus] = list(self.system.get_components(DistributionBus))
        for bus in buses:
            voltage_bases.append(bus.rated_voltage.to("kilovolt").magnitude * 1.732)
        return list(set(voltage_bases))

    def write(  # noqa
        self,
        output_path: Path = Path("./"),
        separate_substations: bool = True,
        separate_feeders: bool = True,
    ):
        base_redirect = set()
        feeders_redirect = defaultdict(set)
        substations_redirect = defaultdict(set)

        self.prepare_folder(output_path)
        component_types = self.system.get_component_types()

        seen_equipment = set()
        seen_controller = set()
        for component_type in component_types:
            # Example component_type is DistributionBus
            components = self.system.get_components(component_type)

            mapper_name = component_type.__name__ + "Mapper"
            # Example mapper_name is string DistributionBusMapper
            if not hasattr(opendss_mapper, mapper_name):
                logger.warning(f"Mapper {mapper_name} not found. Skipping")
                continue

            logger.debug(f"Mapping components in {mapper_name}...")
            mapper = getattr(opendss_mapper, mapper_name)

            # Example mapper is class DistributionBusMapper
            for model in components:
                # Example model is instance of DistributionBus
                if not isinstance(model, DistributionComponentBase) and not (
                    isinstance(model, BareConductorEquipment)
                    or isinstance(model, ConcentricCableEquipment)
                ):
                    continue

                model_map = mapper(model)
                model_map.populate_opendss_dictionary()

                dss_string = self._get_dss_string(model_map)
                if dss_string.startswith("new Vsource"):
                    dss_string = dss_string.replace("new Vsource", "Clear\n\nNew Circuit")
                equipment_dss_string = None
                equipment_map: list[Path] = None
                controller_dss_string = None
                controller_map: list[Path] = None

                if hasattr(model, "equipment"):
                    equipment_mapper_name = model.equipment.__class__.__name__ + "Mapper"
                    if not hasattr(opendss_mapper, equipment_mapper_name):
                        logger.warning(
                            f"Equipment Mapper {equipment_mapper_name} not found. Skipping"
                        )
                    else:
                        equipment_mapper = getattr(opendss_mapper, equipment_mapper_name)
                        equipment_map = equipment_mapper(model.equipment)
                        equipment_map.populate_opendss_dictionary()
                        equipment_dss_string = self._get_dss_string(equipment_map)

                if hasattr(model, "controllers"):
                    for controller in model.controllers:
                        controller_mapper_name = controller.__class__.__name__ + "Mapper"
                        if not hasattr(opendss_mapper, controller_mapper_name):
                            logger.warning(
                                f"Equipment Mapper {controller_mapper_name} not found. Skipping"
                            )
                        else:
                            controller_mapper = getattr(opendss_mapper, controller_mapper_name)
                            controller_map = controller_mapper(controller, model.name)
                            controller_map.populate_opendss_dictionary()
                            controller_dss_string = self._get_dss_string(controller_map)

                output_folder = output_path
                output_redirect = Path("")

                self._build_directory_structure(
                    separate_substations,
                    separate_feeders,
                    output_path,
                    model_map,
                    output_redirect,
                    output_folder,
                )

                if equipment_dss_string is not None:
                    feeder_substation_equipment = (
                        model_map.substation + model_map.feeder + equipment_dss_string
                    )
                    if feeder_substation_equipment not in seen_equipment:
                        seen_equipment.add(feeder_substation_equipment)
                        with open(
                            output_folder / equipment_map.opendss_file, "a", encoding="utf-8"
                        ) as fp:
                            fp.write(equipment_dss_string)

                if controller_dss_string is not None:
                    feeder_substation_controller = (
                        model_map.substation + model_map.feeder + controller_dss_string
                    )
                    if feeder_substation_controller not in seen_controller:
                        seen_controller.add(feeder_substation_controller)
                        with open(
                            output_folder / controller_map.opendss_file, "a", encoding="utf-8"
                        ) as fp:
                            fp.write(controller_dss_string)

                # TODO: Check that there aren't multiple voltage sources for the same master file
                with open(output_folder / model_map.opendss_file, "a", encoding="utf-8") as fp:
                    fp.write(dss_string)

                if (
                    model_map.opendss_file == OpenDSSFileTypes.MASTER_FILE.value
                    or model_map.opendss_file == OpenDSSFileTypes.COORDINATE_FILE.value
                ):
                    continue

                if separate_substations and separate_feeders:
                    substations_redirect[model_map.substation].add(
                        Path(model_map.feeder) / model_map.opendss_file
                    )
                    if equipment_map is not None:
                        substations_redirect[model_map.substation].add(
                            Path(model_map.feeder) / equipment_map.opendss_file
                        )

                elif separate_substations:
                    substations_redirect[model_map.substation].add(Path(model_map.opendss_file))
                    if equipment_map is not None:
                        substations_redirect[model_map.substation].add(
                            Path(equipment_map.opendss_file)
                        )

                if separate_feeders:
                    combined_feeder_sub = Path(model_map.substation) / Path(model_map.feeder)
                    if combined_feeder_sub not in feeders_redirect:
                        feeders_redirect[combined_feeder_sub] = set()
                    feeders_redirect[combined_feeder_sub].add(Path(model_map.opendss_file))
                    if equipment_map is not None:
                        feeders_redirect[combined_feeder_sub].add(Path(equipment_map.opendss_file))

                base_redirect.add(output_redirect / model_map.opendss_file)
                if equipment_map is not None:
                    base_redirect.add(output_redirect / equipment_map.opendss_file)
                if controller_map is not None:
                    base_redirect.add(output_redirect / controller_map.opendss_file)

        self._write_base_master(base_redirect, output_folder)
        self._write_substation_master(substations_redirect)
        self._write_feeder_master(feeders_redirect)

    def _build_directory_structure(
        self,
        separate_substations,
        separate_feeders,
        output_path,
        model_map,
        output_redirect,
        output_folder,
    ):
        if separate_substations:
            output_folder = Path(output_path, model_map.substation)
            output_redirect = Path(model_map.substation)
            output_folder.mkdir(exist_ok=True)
        else:
            output_folder.mkdir(exist_ok=True)

        if separate_feeders:
            output_folder /= model_map.feeder
            output_redirect /= model_map.feeder
            output_folder.mkdir(exist_ok=True)

    def _write_switch_status(self, file_handler: TextIOWrapper):
        switches: list[MatrixImpedanceSwitch] = list(
            self.system.get_components(MatrixImpedanceSwitch)
        )
        for switch in switches:
            if not switch.is_closed[0]:
                file_handler.write(f"open line.{switch.name}\n")

    def _write_base_master(self, base_redirect, output_folder):
        # Only use Masters that have a voltage source, and hence already written.
        sources = list(self.system.get_components(DistributionVoltageSource))
        has_source = True if sources else False

        if has_source:
            bus = self.system.get_source_bus()

            equipment = self.system.get_bus_connected_components(bus.name, DistributionTransformer)
            if equipment:
                equipment_type = "Transformer"
                equipment_name = equipment[0].name
            else:
                equipment = self.system.get_bus_connected_components(
                    bus.name, DistributionBranchBase
                )
                if equipment:
                    equipment_type = "Line"
                    equipment_name = equipment[0].name
                else:
                    equipment_type = None
                    equipment_name = None

        file_order = [file_type.value for file_type in OpenDSSFileTypes]
        master_file = output_folder / OpenDSSFileTypes.MASTER_FILE.value
        if master_file.is_file():
            master_file = output_folder / OpenDSSFileTypes.MASTER_FILE.value
            with open(master_file, "a") as base_master:
                # TODO: provide ordering so LineCodes before Lines
                for file in file_order:
                    for dss_file in base_redirect:
                        if dss_file.name == file:
                            if (master_file.parent / dss_file).exists():
                                base_master.write("redirect " + str(dss_file))
                                base_master.write("\n")
                                break
                self._write_switch_status(base_master)

                if has_source and equipment_type:
                    base_master.write(
                        f"New EnergyMeter.SourceMeter element={equipment_type}.{equipment_name}\n"
                    )
                base_master.write(f"Set Voltagebases={self._get_voltage_bases()}\n")
                base_master.write("calcv\n")
                base_master.write("Solve\n")
                base_master.write(f"redirect {OpenDSSFileTypes.COORDINATE_FILE.value}\n")

        # base_master.write(f"BusCoords {filename}\n")

    def _write_substation_master(self, substations_redirect):
        for substation in substations_redirect:
            if (Path(substation) / OpenDSSFileTypes.MASTER_FILE.value).is_file():
                with open(
                    Path(substation) / OpenDSSFileTypes.MASTER_FILE.value, "a"
                ) as substation_master:
                    # TODO: provide ordering so LineCodes before Lines
                    for dss_file in substations_redirect[substation]:
                        if (Path(substation).parent / dss_file).exists():
                            substation_master.write("redirect " + str(dss_file))
                            substation_master.write("\n")

    def _write_feeder_master(self, feeders_redirect):
        for feeder in feeders_redirect:
            if (Path(feeder) / OpenDSSFileTypes.MASTER_FILE.value).is_file():
                with open(Path(feeder) / OpenDSSFileTypes.MASTER_FILE.value, "a") as feeder_master:
                    # TODO: provide ordering so LineCodes before Lines
                    for dss_file in feeders_redirect[feeder]:
                        if (Path(feeder).parent / dss_file).exists():
                            feeder_master.write("redirect " + str(dss_file))
                            feeder_master.write("\n")



from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class SequenceImpedanceBranchEquipmentMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "LineCode_Z0Z1C0C1"
    altdss_composition_name = "LineCode"
    opendss_file = OpenDSSFileTypes.LINECODES_FILE.value

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_common(self):
        self.opendss_dict["Units"] = "km"

    # TODO: Do we need to add NPhases? This might be very important for secondary lines.

    def map_pos_seq_resistance(self):
        pos_seq_resistance_ohms = self.model.pos_seq_resistance.to("ohm/km")
        self.opendss_dict["R1"] = pos_seq_resistance_ohms.magnitude

    def map_zero_seq_resistance(self):
        zero_seq_resistance_ohms = self.model.zero_seq_resistance.to("ohm/km")
        self.opendss_dict["R0"] = zero_seq_resistance_ohms.magnitude

    def map_pos_seq_reactance(self):
        pos_seq_reactance_ohms = self.model.pos_seq_reactance.to("ohm/km")
        self.opendss_dict["X1"] = pos_seq_reactance_ohms.magnitude

    def map_zero_seq_reactance(self):
        zero_seq_reactance_ohms = self.model.zero_seq_reactance.to("ohm/km")
        self.opendss_dict["X0"] = zero_seq_reactance_ohms.magnitude

    def map_pos_seq_capacitance(self):
        pos_seq_capacitance_nf = self.model.pos_seq_capacitance.to("nanofarad/km")
        self.opendss_dict["C1"] = pos_seq_capacitance_nf.magnitude

    def map_zero_seq_capacitance(self):
        zero_seq_capacitance_nf = self.model.zero_seq_capacitance.to("nanofarad/km")
        self.opendss_dict["C0"] = zero_seq_capacitance_nf.magnitude

    def map_ampacity(self):
        ampacity_amps = self.model.ampacity.to("ampere")
        self.opendss_dict["NormAmps"] = ampacity_amps.magnitude



from gdm.distribution.enums import VoltageTypes, ConnectionType, Phase
from gdm.quantities import Voltage, ApparentPower
from gdm.distribution.equipment import WindingEquipment

from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper
from ditto.readers.cim_iec_61968_13.common import phase_mapper


class WindingEquipmentMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        indices = row.index
        windings = []
        for i in range(1, 3):
            new_indices = [j for j in indices if j.startswith(f"wdg_{i}_")]
            windings.append(self._build_winding(row[new_indices], i, row["xfmr"]))

        return windings

    def _build_winding(self, row, index, xfmr_name):
        if f"wdg_{index}_phase" in row:
            self.phases = [phase_mapper[phs] for phs in row[f"wdg_{index}_phase"].replace("N", "")]
        else:
            self.phases = [Phase.A, Phase.B, Phase.C]
        self.n_phases = len(self.phases)

        mapping = {
            f"wdg_{index}_normal_tap": ["normal_tap", 0],
            f"wdg_{index}_max_tap": ["max_tap", 16],
            f"wdg_{index}_min_tap": ["min_tap", -16],
            f"wdg_{index}_dv": ["dv", 0.625],
        }

        for k, v in mapping.items():
            if k in row:
                if row[k]:
                    setattr(self, v[0], float(row[k]))
                else:
                    setattr(self, v[0], v[1])
            else:
                setattr(self, v[0], v[1])

        self.dv = self.dv / 100.0
        self.total_taps = self.max_tap - self.min_tap
        self.pu_tap = self.normal_tap * self.dv + 1
        self.max_tap_pu = self.max_tap * self.dv + 1
        self.min_tap_pu = self.min_tap * self.dv + 1

        return WindingEquipment(
            name=self.map_name(index, xfmr_name),
            resistance=self.map_resistance(row, index),
            is_grounded=self.map_is_grounded(row, index),
            rated_voltage=self.map_rated_voltage(row, index),
            voltage_type=self.map_voltage_type(row, index),
            rated_power=self.map_rated_power(row, index),
            num_phases=self.map_num_phases(row, index),
            connection_type=self.map_connection_type(row, index),
            tap_positions=self.map_tap_positions(row, index),
            total_taps=self.map_total_taps(row, index),
            min_tap_pu=self.map_min_tap_pu(row, index),
            max_tap_pu=self.map_max_tap_pu(row, index),
        )

    def map_name(self, winding_number, xfmr_name):
        return f"{xfmr_name}_winding_{winding_number}"

    def map_resistance(self, row, winding_number):
        s = float(row[f"wdg_{winding_number}_apparent_power"])
        v = float(row[f"wdg_{winding_number}_rated_voltage"])
        r = float(row[f"wdg_{winding_number}_per_resistance"])
        per_r = r / (v**2 / s) * 100
        return per_r

    def map_is_grounded(self, row, winding_number):
        return False

    def map_rated_voltage(self, row, winding_number):
        voltage = float(row[f"wdg_{winding_number}_rated_voltage"])
        if self.n_phases > 1:
            return Voltage(voltage / 1.732, "volt")
        else:
            return Voltage(voltage, "volt")

    def map_voltage_type(self, row, winding_number):
        return VoltageTypes.LINE_TO_GROUND

    def map_rated_power(self, row, winding_number):
        rated_power = float(row[f"wdg_{winding_number}_apparent_power"])
        return ApparentPower(rated_power, "va")

    def map_num_phases(self, row, winding_number):
        return self.n_phases

    def map_connection_type(self, row, winding_number):
        connection = row[f"wdg_{winding_number}_conn"]

        if connection == "D":
            return ConnectionType.DELTA
        else:
            return ConnectionType.STAR

    # TODO: Check if this is correct
    def map_tap_positions(self, row, winding_number):
        return [self.pu_tap] * self.n_phases

    def map_total_taps(self, row, winding_number):
        return self.total_taps

    def map_min_tap_pu(self, row, winding_number):
        return self.min_tap_pu

    def map_max_tap_pu(self, row, winding_number):
        return self.max_tap_pu



length_units = {
    "English2": {
        "SUL": "inch",
        "MUL": "feet",
        "LUL": "mile",
        "PerLUL": "ohm/mile",
    },
    "English1": {
        "SUL": "inch",
        "MUL": "feet",
        "LUL": "kft",
        "PerLUL": "ohm/kft",
    },
    "English": {
        "SUL": "inch",
        "MUL": "feet",
        "LUL": "kft",
        "PerLUL": "ohm/kft",
    },
    "Metric": {
        "SUL": "mm",
        "MUL": "m",
        "LUL": "km",
        "PerLUL": "ohm/km",
    },
}



from abc import ABC, abstractmethod
from infrasys import Component


# TODO: Define a BaseMapper class one level up from this?
class OpenDSSMapper(ABC):
    phase_map = {"A": ".1", "B": ".2", "C": ".3", "N": ".0", "S1": ".1", "S2": ".2"}
    length_units_map = {
        "meter": "m",
        "mile": "mi",
        "feet": "ft",
        "foot": "ft",
        "kilometer": "km",
        "inch": "in",
        "centimeter": "cm",
        "millimeter": "mm",
    }
    connection_map = {"STAR": "wye", "DELTA": "delta", "OPEN_DELTA": "delta", "OPEN_STAR": "wye"}

    def __init__(self, model):
        self.model: Component = model
        self.opendss_dict = {}
        self.substation = ""
        self.feeder = ""

        @property
        @abstractmethod
        def opendss_file():
            """Return the OpenDSS file."""
            pass

        @property
        @abstractmethod
        def altdss_name():
            """Return the name of the AltDSS class which defines the object."""
            pass

        @property
        @abstractmethod
        def altdss_composition_name():
            """Return the name of the AltDSS class which constructs the object through composition"""
            pass

    def map_common(self):
        return

    def map_uuid(self):
        return

    def map_system_uuid(self):
        return

    def map_substation(self):
        if hasattr(self.model, "substation") and self.model.substation is not None:
            self.substation = self.model.substation.name

    def map_feeder(self):
        if hasattr(self.model, "feeder") and self.model.feeder is not None:
            self.feeder = self.model.feeder.name

    def populate_opendss_dictionary(self):
        # Should not be populating an existing dictionary. Assert error if not empty
        assert len(self.opendss_dict) == 0
        self.map_common()
        for field in self.model.model_fields:
            mapping_function = getattr(self, "map_" + field)
            mapping_function()



from gdm.distribution.enums import ConnectionType

from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class DistributionCapacitorMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "Capacitor_kvarkV"
    altdss_composition_name = "Capacitor"
    opendss_file = OpenDSSFileTypes.CAPACITORS_FILE.value

    def map_in_service(self):
        self.opendss_dict["enabled"] = self.model.in_service

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_bus(self):
        self.opendss_dict["Bus1"] = self.model.bus.name
        num_phases = len(self.model.phases)
        for phase in self.model.phases:
            self.opendss_dict["Bus1"] += self.phase_map[phase]
        # TODO: Should we include the phases its connected to here?
        nom_voltage = self.model.bus.rated_voltage.to("kV").magnitude
        self.opendss_dict["kV"] = nom_voltage if num_phases == 1 else nom_voltage * 1.732

    def map_phases(self):
        if (
            len(self.model.phases) == 2
            and self.model.equipment.connection_type == ConnectionType.DELTA
        ):
            self.opendss_dict["Phases"] = 1
        else:
            self.opendss_dict["Phases"] = len(self.model.phases)
        # TODO: Do we need to remove neutrals?

    def map_controllers(self):
        # controller = self.model.controllers
        # TODO: The controller isn't included in the capacitor mapping.
        ...

    def map_equipment(self):
        equipment = self.model.equipment
        connection = self.connection_map[equipment.connection_type]
        self.opendss_dict["Conn"] = connection
        total_resistance = []
        total_reactance = []
        total_rated_reactive_power = []
        # TODO: Note that this sets the NumSteps to be the number of phase capacitors. Is this right? Do we need to check if banked or not?
        num_banks = None
        for phase_capacitor in equipment.phase_capacitors:
            num_banks = phase_capacitor.num_banks
            total_resistance.append(phase_capacitor.resistance.to("ohm").magnitude)
            total_reactance.append(phase_capacitor.reactance.to("ohm").magnitude)
            total_rated_reactive_power.append(
                phase_capacitor.rated_reactive_power.to("kvar").magnitude
            )  # from general capacitor equipment
            self.opendss_dict["States"] = [1] * num_banks
        self.opendss_dict["R"] = [sum(total_resistance) / num_banks] * num_banks
        self.opendss_dict["XL"] = [sum(total_reactance) / num_banks] * num_banks
        total_kvar_per_bank = sum(total_rated_reactive_power) / num_banks
        self.opendss_dict["kvar"] = [total_kvar_per_bank] * num_banks

        # TODO: We're not building equipment for the Capacitors. This means that there's no guarantee that we're addressing all of the attributes in the equipment in a structured way like we are for the component.



from gdm.quantities import (
    ResistancePULength,
    Distance,
    Current,
    Voltage,
)
from gdm.distribution.equipment import ConcentricCableEquipment
from pydantic import PositiveInt
import opendssdirect as odd
from loguru import logger

from ditto.readers.opendss.common import query_model_data, get_equipment_from_catalog


def get_cables_equipment() -> list[ConcentricCableEquipment]:
    """Method returns a list of ConcentricCableEquipment objects

    Returns:
        list[ConcentricCableEquipment]: list of ConcentricCableEquipment
    """

    logger.debug("parsing cable components...")
    concentric_cable_equipment_catalog = {}
    model_type = "CNData"
    cables = []
    odd.Circuit.SetActiveClass(model_type)
    flag = odd.ActiveClass.First()
    while flag > 0:
        model_name = odd.Element.Name().lower().split(".")[1]
        gmr_units = query_model_data(model_type, model_name, "gmrunits", str)
        radius_units = query_model_data(model_type, model_name, "radunits", str)
        length_units = query_model_data(model_type, model_name, "runits", str)

        gmr = query_model_data(model_type, model_name, "gmr", float)
        diam = query_model_data(model_type, model_name, "diam", float)
        gmr_strand = query_model_data(model_type, model_name, "gmr", float)
        diam_strand = query_model_data(model_type, model_name, "diam", float)

        cable = ConcentricCableEquipment.model_construct(
            strand_ac_resistance=ResistancePULength(
                query_model_data(model_type, model_name, "rstrand", float), f"ohms/{length_units}"
            ),
            dc_resistance=ResistancePULength(
                query_model_data(model_type, model_name, "rdc", float), f"ohms/{length_units}"
            ),
            phase_ac_resistance=ResistancePULength(
                query_model_data(model_type, model_name, "rac", float), f"ohms/{length_units}"
            ),
            strand_gmr=Distance(
                gmr_strand if gmr_strand else diam_strand * 0.7788, f"{radius_units}"
            ),
            strand_diameter=Distance(
                diam_strand if diam_strand else gmr_strand / 0.7788, f"{radius_units}"
            ),
            ampacity=Current(
                query_model_data(model_type, model_name, "normamps", float), "ampere"
            ),
            emergency_ampacity=Current(
                query_model_data(model_type, model_name, "emergamps", float), "ampere"
            ),
            insulation_thickness=Distance(
                query_model_data(model_type, model_name, "InsLayer", float), "ampere"
            ),
            cable_diameter=Distance(
                query_model_data(model_type, model_name, "DiaCable", float), "ampere"
            ),
            insulation_diameter=Distance(
                query_model_data(model_type, model_name, "diains", float), "ampere"
            ),
            num_neutral_strands=PositiveInt(
                query_model_data(model_type, model_name, "k", int), "ampere"
            ),
            conductor_diameter=Distance(diam if diam else gmr / 0.7788, f"{radius_units}"),
            conductor_gmr=Distance(gmr if gmr else diam * 0.7788, f"{gmr_units}"),
            rated_voltage=Voltage(12.47, "volts"),
            name=model_type,
        )

        cable = get_equipment_from_catalog(cable, concentric_cable_equipment_catalog)

        cables.append(cable)
        flag = odd.ActiveClass.Next()
    return cables



from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class DistributionTransformerMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "Transformer_XfmrCode"
    altdss_composition_name = "Transformer"
    opendss_file = OpenDSSFileTypes.TRANSFORMERS_FILE.value

    def map_in_service(self):
        self.opendss_dict["enabled"] = self.model.in_service

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_buses(self):
        buses = []
        phases = []
        buses_and_phases = []
        is_center_tapped = self.model.equipment.is_center_tapped
        if is_center_tapped:
            for i in range(len(self.model.buses)):
                bus = self.model.buses[i]
                buses.append(bus.name)
            dss_phases = ""
            for phase in self.model.winding_phases[0]:
                dss_phases += self.phase_map[phase]
            phases.append(dss_phases)
            phases.append(".1.0")
            phases.append(".0.2")

        else:
            for bus in self.model.buses:
                buses.append(bus.name)
            for winding_phases in self.model.winding_phases:
                dss_phases = ""
                for phase in winding_phases:
                    dss_phases += self.phase_map[phase]
                phases.append(dss_phases)

        for i in range(len(buses)):
            buses_and_phases.append(buses[i] + phases[i])
        self.opendss_dict["Bus"] = buses_and_phases

    def map_winding_phases(self):
        primary_winding = self.model.winding_phases[0]
        self.opendss_dict["Phases"] = len(primary_winding)

    def map_equipment(self):
        equipment = self.model.equipment
        self.opendss_dict["XfmrCode"] = equipment.name



from gdm.distribution.components import DistributionBus, DistributionCapacitor

from ditto.readers.cim_iec_61968_13.equipment.capacitor_equipment import CapacitorEquipmentMapper
from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper
from ditto.readers.cim_iec_61968_13.common import phase_mapper


class DistributionCapacitorMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        return DistributionCapacitor(
            name=self.map_name(row),
            bus=self.map_bus(row),
            phases=self.map_phases(row),
            controllers=self.map_controllers(row),
            equipment=self.map_equipment(row),
        )

    def map_name(self, row):
        return row["capacitor"]

    def map_bus(self, row):
        bus_name = row["bus"]
        bus = self.system.get_component(component_type=DistributionBus, name=bus_name)
        return bus

    def map_phases(self, row):
        phases = row["phase"]
        if phases is None:
            phases = ["A", "B", "C"]
        else:
            phases = phases.split(",")
        return [phase_mapper[phase] for phase in phases]

    def map_controllers(self, row):
        return []

    def map_equipment(self, row):
        mapper = CapacitorEquipmentMapper(self.system)
        equipment = mapper.parse(row)
        return equipment



from enum import IntEnum
from typing import Any
import ast

from gdm.distribution.components import DistributionVoltageSource
from gdm.distribution.enums import Phase
from infrasys import Component, System
import opendssdirect as odd


PHASE_MAPPER = {
    "0": Phase.N,
    "1": Phase.A,
    "2": Phase.B,
    "3": Phase.C,
    "4": Phase.N,
}

UNIT_MAPPER = {0: "m", 1: "mi", 2: "kft", 3: "km", 4: "m", 5: "ft", 6: "in", 7: "cm"}


class LoadTypes(IntEnum):
    """Load types represented in Ditto"""

    CONST_POWER = 1
    CONST_IMPEDANCE = 2
    CONST_P__QUARDRATIC_Q = 3
    LINEAR_P__QUARDRATIC_Q = 4
    CONST_CURRENT = 5
    ZIP = 8


def hash_model(model: Component, key_names: list[str] = ["name", "uuid"]) -> int:
    """Return hash of the passed model

    Args:
        model (Component): Instance of a derived infrasys Component model
        key_names (list[str], optional): List of keys to be removed from the model. Defaults to ["name", "uuid"].

    Returns:
        int: model hash
    """
    model_dict = (
        model.model_dump()
    )  # TODO: exclude={"name"} seems not to work  well with list of objects
    cleaned_model = remove_keys_from_dict(model_dict, key_names)
    return hash(str(cleaned_model))


def remove_keys_from_dict(model_dict: dict, key_names: list[str] = ["name", "uuid"]) -> dict:
    """Method recursively removes keys from the model

    Args:
        model_dict (dict): model in dict representation
        key_names (list[str]): keys to remove from the model

    Returns:
        dict: reduced model dictionary
    """
    for key_name in key_names:
        if key_name in model_dict:
            model_dict.pop(key_name)
        for k, v in model_dict.items():
            if isinstance(v, dict):
                model_dict[k] = remove_keys_from_dict(v)
            elif isinstance(v, list):
                values = []
                for value in v:
                    if isinstance(value, dict):
                        value = remove_keys_from_dict(value)
                    values.append(value)
                    model_dict[k] = values
    return model_dict


def get_equipment_from_catalog(
    model: Component, catalog: dict[int, Component], sub_catalog: str | None = None
) -> Component | None:
    """If the equipment already exixts in th system the equipment instance is returned else None is returned

    Args:
        model (Component): Instance of GDM equipment
        catalog (dict[int, Component]): mapping model hash to model
        sub_catalog (str | None, optional): sub catalog in a catalog. Defaults to None.

    Returns:
        Component | None:  Instance of GDM equipment
    """

    model_hash = hash_model(model)
    if sub_catalog is None:
        if model_hash in catalog:
            return catalog[model_hash]
        else:
            catalog[model_hash] = model
            return model
    else:
        assert sub_catalog in catalog and isinstance(catalog[sub_catalog], dict)
        if model_hash in catalog[sub_catalog]:
            return catalog[sub_catalog][model_hash]
        else:
            catalog[sub_catalog][model_hash] = model
            return model


def query_model_data(model_type: str, model_name: str, property: str, dtype: type) -> Any:
    """query OpenDSS model property

    Args:
        model_type (str): OpenDSS model type
        model_name (str): OpenDSS model name
        property (str): OpenDSS model property
        dtype (type): data type e.g. str, float

    Returns:
        Any: OpenDSS model property value
    """
    command = f"? {model_type}.{model_name}.{property}"
    odd.Text.Command(command)
    result = odd.Text.Result()
    if result == "Property Unknown":
        return None
    if dtype is list:
        result = result.replace("[ ", "[").replace(" ", ",")
        return ast.literal_eval(result)
    else:
        return dtype(result)


def get_source_bus(system: System) -> str:
    """Returns the name of the source bus

    Args:
        system (System): Instance of an Infrasys System
    """

    voltage_sources = system.get_components(DistributionVoltageSource)
    buses = [v_source.bus.name for v_source in voltage_sources]
    return buses



from uuid import uuid4

from gdm.distribution.enums import VoltageTypes
from gdm.quantities import Voltage
from gdm.distribution.components import DistributionSolar, DistributionBus
from gdm.distribution.equipment import (
    InverterEquipment,
    SolarEquipment,
)
from gdm.quantities import (
    ApparentPower,
    ActivePower,
    ReactivePower,
    Irradiance,
)


from infrasys.system import System
import opendssdirect as odd
from loguru import logger

from ditto.readers.opendss.common import PHASE_MAPPER, get_equipment_from_catalog
from ditto.readers.opendss.components.loadshapes import build_profiles, ObjectsWithProfile


def _build_pv_equipment(
    solar_equipment_catalog: dict[int, SolarEquipment],
) -> tuple[SolarEquipment, list[str], str, list[str]]:
    """Helper function to build a SolarEquipment instance

    Args:
        solar_equipment_catalog (dict[int, SolarEquipment]): mapping of model hash to SolarEquipment instance

    Returns:
        SolarEquipment: instance of SolarEquipment
        list[str]: List of buses
        list[str]: List of phases
    """

    logger.debug("parsing pvsystem equipment...")
    pv_name = odd.PVsystems.Name()

    def query(ppty):
        odd.Text.Command(f"? pvsystem.{pv_name}.{ppty}")
        return odd.Text.Result()

    equipment_uuid = uuid4()
    buses = odd.CktElement.BusNames()
    num_phase = odd.CktElement.NumPhases()
    # kva_ac = odd.PVsystems.kVARated()
    kw_dc = odd.PVsystems.Pmpp()
    nodes = buses[0].split(".")[1:] if num_phase != 3 else ["1", "2", "3"]
    live_nodes = [node for node in nodes if node != "0"]

    voltage_type = (
        VoltageTypes.LINE_TO_GROUND
        if num_phase == 1 and len(live_nodes)
        else VoltageTypes.LINE_TO_LINE
    )

    solar_equipment = SolarEquipment.model_construct(
        name=str(equipment_uuid),
        rated_power=ActivePower(kw_dc, "kilova"),
        resistance=float(query(r"%r")),
        reactance=float(query(r"%x")),
        rated_voltage=Voltage(float(query(r"kv")), "kilovolt"),
        voltage_type=voltage_type,
    )
    solar_equipment = get_equipment_from_catalog(solar_equipment, solar_equipment_catalog)
    return solar_equipment, buses, nodes


def get_pvsystems(system: System) -> list[DistributionSolar]:
    """Function to return list of DistributionSolar in Opendss model.

    Args:
        system (System): Instance of System

    Returns:
        List[DistributionSolar]: List of DistributionSolar objects
    """

    logger.debug("parsing pvsystem components...")
    solar_equipment_catalog = {}
    profile_catalog = {}
    pv_systems = []
    flag = odd.PVsystems.First()
    while flag > 0:
        logger.debug(f"building pvsystem {odd.PVsystems.Name()}...")
        solar_name = odd.PVsystems.Name().lower()

        def query(ppty):
            odd.Text.Command(f"? pvsystem.{solar_name}.{ppty}")
            return odd.Text.Result()

        solar_equipment, buses, nodes = _build_pv_equipment(solar_equipment_catalog)
        bus1 = buses[0].split(".")[0]
        distribution_solar = DistributionSolar.model_construct(
            name=solar_name,
            bus=system.get_component(DistributionBus, bus1),
            phases=[PHASE_MAPPER[el] for el in nodes],
            irradiance=Irradiance(odd.PVsystems.Irradiance(), "kilowatt/meter**2"),
            active_power=ActivePower(odd.PVsystems.kW(), "kilowatt"),
            reactive_power=ReactivePower(odd.PVsystems.kvar(), "kilovar"),
            inverter=InverterEquipment.model_construct(
                name=str(uuid4()),
                rated_apparent_power=ApparentPower(odd.PVsystems.kVARated(), "kilova"),
                rise_limit=None,
                fall_limit=None,
                eff_curve=None,
                cutout_percent=float(query(r"%cutout")),
                cutin_percent=float(query(r"%cutin")),
                dc_to_ac_efficiency=float(query(r"%pmpp")),
            ),
            controller=None,
            equipment=solar_equipment,
        )
        profile_names = [odd.PVsystems.daily(), odd.PVsystems.yearly(), odd.PVsystems.duty()]
        profiles = build_profiles(profile_names, ObjectsWithProfile.PV_SYSTEM, profile_catalog)

        for profile_name in profile_names:
            if profile_name in profiles:
                for profile_type, ts_data in profiles[profile_name].items():
                    system.add_time_series(
                        ts_data["data"],
                        distribution_solar,
                        profile_type=profile_type,
                        profile_name=profile_name,
                        use_actual=ts_data["use_actual"],
                    )
                    logger.debug(
                        f"Adding timeseries profile '{profile_name} / {profile_type}' to solar '{solar_name}'"
                    )
        pv_systems.append(distribution_solar)
        flag = odd.PVsystems.Next()
    return pv_systems



from gdm.distribution.equipment import PhaseLoadEquipment, LoadEquipment
from gdm.distribution.components import DistributionBus
from gdm.quantities import ActivePower, ReactivePower
from gdm.distribution.enums import ConnectionType

from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper
from ditto.readers.cim_iec_61968_13.common import phase_mapper


class LoadEquipmentMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        phases = row["phase"]
        if phases is None:
            self.phases = ["A", "B", "C"]
        else:
            self.phases = phases.split(",")

        phases = [phase_mapper[phase] for phase in self.phases]
        bus = self.system.get_component(component_type=DistributionBus, name=row["bus"])

        if row["grounded"] == "false" and len(phases) == 1:
            diff = list(set(bus.phases).difference(phases))
            if diff:
                self.phases.append(diff[-1].value)

        return LoadEquipment(
            name=self.map_name(row),
            phase_loads=self.map_phase_loads(row),
            connection_type=self.map_connection_type(row),
        )

    # NOTE: Names may not be unique. Should we append a number to the name?
    def map_name(self, row):
        return row["load"] + "_equipment"

    # No connection type information is included
    def map_connection_type(self, row):
        return ConnectionType.DELTA if row["conn"] == "D" else ConnectionType.STAR

    def map_phase_loads(self, row):
        phase_loads = []
        n_phases = len(self.phases)
        kw = float(row["active power"]) / n_phases
        kvar = float(row["reactive power"]) / n_phases
        for phase in self.phases:
            if kw > 0:
                mapper = PhaseLoadEquipmentMapper(self.system)
                phase_load = mapper.parse(row, kw, kvar, phase)
                phase_loads.append(phase_load)
        return phase_loads


class PhaseLoadEquipmentMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row, kw, kvar, phase):
        self.kw = kw
        self.kvar = kvar
        return PhaseLoadEquipment(
            name=self.map_name(row, phase),
            real_power=self.map_real_power(kw),
            reactive_power=self.map_reactive_power(kvar),
            z_real=self.map_z_real(row),
            z_imag=self.map_z_imag(row),
            i_real=self.map_i_real(row),
            i_imag=self.map_i_imag(row),
            p_real=self.map_p_real(row),
            p_imag=self.map_p_imag(row),
            num_customers=self.map_num_customers(row),
        )

    def map_name(self, row, phase):
        return row["load"] + "_phase_load_equipment_" + phase

    def map_real_power(self, kw):
        return ActivePower(kw, "watt")

    def map_reactive_power(self, kvar):
        return ReactivePower(kvar, "var")

    def map_z_real(self, row):
        return float(row["z_p"]) / 100.0

    def map_z_imag(self, row):
        return float(row["z_q"]) / 100.0

    def map_i_real(self, row):
        return float(row["i_p"]) / 100.0

    def map_i_imag(self, row):
        return float(row["i_q"]) / 100.0

    def map_p_real(self, row):
        return float(row["p_p"]) / 100.0

    def map_p_imag(self, row):
        return float(row["p_q"]) / 100.0

    def map_num_customers(self, row):
        return None



from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes

from gdm.distribution.controllers import RegulatorController


class RegulatorControllerMapper(OpenDSSMapper):
    altdss_name = "RegControl"
    # altdss_composition_name = "RegControl"
    altdss_composition_name = None
    opendss_file = OpenDSSFileTypes.REGULATOR_CONTROLLERS_FILE.value

    def __init__(self, model: RegulatorController, xfmr_name: str):
        super().__init__(model)
        self.model: RegulatorController = model
        self.xfmr_name = xfmr_name

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name
        self.opendss_dict["Transformer"] = self.xfmr_name

    def map_delay(self):
        self.opendss_dict["TapDelay"] = self.model.delay.to("s").magnitude

    def map_v_setpoint(self):
        self.opendss_dict["VReg"] = self.model.v_setpoint.to("volts").magnitude

    def map_min_v_limit(self):
        self.model.min_v_limit.to("volts").magnitude

    def map_max_v_limit(self):
        self.model.max_v_limit.to("volts").magnitude

    def map_pt_ratio(self):
        self.opendss_dict["PTRatio"] = self.model.pt_ratio

    def map_use_ldc(self):
        ...

    def map_is_reversible(self):
        self.opendss_dict["Reversible"] = self.model.is_reversible

    def map_ldc_R(self):
        self.opendss_dict["R"] = self.model.ldc_R.to("volts").magnitude

    def map_ldc_X(self):
        self.opendss_dict["X"] = self.model.ldc_X.to("volts").magnitude

    def map_ct_primary(self):
        self.opendss_dict["CTPrim"] = self.model.ct_primary.to("ampere").magnitude

    def map_max_step(self):
        self.opendss_dict["MaxTapChange"] = self.model.max_step

    def map_bandwidth(self):
        self.opendss_dict["Band"] = self.model.bandwidth.to("volts").magnitude

    def map_controlled_bus(self):
        self.opendss_dict[
            "Bus"
        ] = f"{self.model.controlled_bus.name}{self.phase_map[self.model.controlled_phase]}"

    def map_controlled_phase(self):
        ...



from pathlib import Path

from gdm.distribution.equipment import MatrixImpedanceBranchEquipment
from gdm.distribution.controllers import RegulatorController
from gdm.distribution import DistributionSystem
from gdm.distribution.components import (
    DistributionComponentBase,
    DistributionVoltageSource,
    DistributionTransformer,
    MatrixImpedanceBranch,
    DistributionCapacitor,
    DistributionRegulator,
    MatrixImpedanceSwitch,
    DistributionLoad,
    DistributionBus,
)
from loguru import logger
from rdflib import Graph
import pandas as pd


from ditto.readers.cim_iec_61968_13.queries import (
    query_distribution_regulators,
    query_regulator_controllers,
    query_transformer_windings,
    query_load_break_switches,
    query_power_transformers,
    query_distribution_buses,
    query_line_segments,
    query_line_codes,
    query_capacitors,
    query_source,
    query_loads,
)
import ditto.readers.cim_iec_61968_13 as cim_mapper
from ditto.readers.reader import AbstractReader


class Reader(AbstractReader):
    # NOTE:  Do not change sequnce of the component types below.
    component_types: DistributionComponentBase = [
        DistributionBus,
        DistributionLoad,
        DistributionCapacitor,
        DistributionVoltageSource,
        RegulatorController,
        MatrixImpedanceBranchEquipment,
        MatrixImpedanceBranch,
        DistributionTransformer,
        DistributionRegulator,
        MatrixImpedanceSwitch,
    ]

    def __init__(self, cim_file: str | Path):
        cim_file = Path(cim_file)
        assert cim_file.exists(), f"{cim_file} does not exist"
        self.system = DistributionSystem(auto_add_composed_components=True)
        self.graph = Graph()
        self.graph.parse(cim_file, format="xml")

    def read(self):
        datasets: dict[DistributionComponentBase, pd.DataFrame] = {}
        logger.debug("Querying for distribution buses...")
        datasets[DistributionBus] = query_distribution_buses(self.graph)

        logger.debug("Querying for AC line segments...")
        datasets[MatrixImpedanceBranch] = query_line_segments(self.graph)

        logger.debug("Querying for line codes...")
        datasets[MatrixImpedanceBranchEquipment] = query_line_codes(self.graph)

        logger.debug("Querying for loads...")
        datasets[DistributionLoad] = query_loads(self.graph)

        logger.debug("Querying for capacitors...")
        datasets[DistributionCapacitor] = query_capacitors(self.graph)

        logger.debug("Querying for transformers...")
        xfmr_data = query_power_transformers(self.graph)
        logger.debug("Querying for transformer windings...")
        winding_data = query_transformer_windings(self.graph)
        datasets[DistributionTransformer] = self._build_xfmr_dataset(xfmr_data, winding_data)

        logger.debug("Querying for regulators...")
        regulator_data = query_distribution_regulators(self.graph)
        datasets[DistributionRegulator] = self._build_xfmr_dataset(regulator_data)

        logger.debug("Querying for sources...")
        datasets[DistributionVoltageSource] = query_source(self.graph)

        logger.debug("Querying for regulator controllers...")
        datasets[RegulatorController] = query_regulator_controllers(self.graph)

        logger.debug("Querying for load break switches...")
        datasets[MatrixImpedanceSwitch] = query_load_break_switches(self.graph)

        datasets[DistributionBus] = self._set_bus_phases(datasets)

        for component_type in self.component_types:
            mapper_name = component_type.__name__ + "Mapper"
            components = []
            if component_type in datasets:
                try:
                    mapper = getattr(cim_mapper, mapper_name)(self.system)
                    logger.debug(f"Buliding components for {component_type.__name__}")
                except AttributeError as _:
                    logger.warning(f"Mapper for {mapper_name} not found. Skipping")
                if datasets[component_type].empty:
                    logger.warning(
                        f"Dataframe for {component_type.__name__} is empty. Check query."
                    )
                for _, row in datasets[component_type].iterrows():
                    model_entry = mapper.parse(row)
                    components.append(model_entry)
            else:
                logger.warning(f"Dataframe for {component_type.__name__} not found. Skipping")
            self.system.add_components(*components)
        logger.info("System summary: ", self.system.info())

    def _build_xfmr_dataset(
        self, xfmr_data: pd.DataFrame, winding_df: pd.DataFrame = pd.DataFrame()
    ) -> pd.DataFrame:
        xfmrs = xfmr_data["xfmr"].unique()
        xfms = []
        for xfmr in xfmrs:
            xfmr_df = xfmr_data[xfmr_data["xfmr"] == xfmr]
            xfmr_df.pop("xfmr")
            windings = xfmr_df.pop("winding").unique()
            buses = xfmr_df.pop("bus").unique()
            xfmr_df = xfmr_df.drop_duplicates()
            wdgs = []
            for winding, (_, winding_data) in zip(windings, xfmr_df.iterrows()):
                winding_data.index = [f"wdg_{winding}_" + c for c in winding_data.index]
                wdgs.append(winding_data)
            wdgs = pd.concat(wdgs)
            wdgs["bus_1"] = buses[0]
            wdgs["bus_2"] = buses[1]
            wdgs["xfmr"] = xfmr
            for _, wdg_coupling_data in winding_df.iterrows():
                xfmr_ends = {wdg_coupling_data["xfmr_end_1"], wdg_coupling_data["xfmr_end_2"]}
                intersection = xfmr_ends.intersection(wdgs.to_list())
                if intersection:
                    wdgs["r0"] = wdg_coupling_data["r0"]
                    wdgs["r1"] = wdg_coupling_data["r1"]
                    wdgs["x0"] = wdg_coupling_data["x0"]
                    wdgs["x1"] = wdg_coupling_data["x1"]
                    wdgs["winding"] = wdg_coupling_data["winding"]
            xfms.append(wdgs)

        xfmr_dataset = pd.DataFrame(xfms)
        return xfmr_dataset

    def _set_bus_phases(
        self, df_dict: dict[DistributionComponentBase, pd.DataFrame]
    ) -> pd.DataFrame:
        all_phases = []
        bus_df = df_dict.pop(DistributionBus)
        for _, bus_data in bus_df.iterrows():
            bus_name = bus_data["bus"]
            phases = []

            for df_type in df_dict:
                df = df_dict[df_type]
                bus_subset = {"bus", "bus_1", "bus_2"}
                phase_subset = {"phase", "phases_1", "phases_2"}

                bus_cols = bus_subset.intersection(df.columns)
                phase_cols = phase_subset.intersection(df.columns)

                if isinstance(df, pd.DataFrame) and len(bus_cols) and len(phase_cols):
                    for bus_col in bus_subset:
                        if bus_col in df.columns:
                            filt_data = df[df[bus_col] == bus_name]
                            if not filt_data.empty:
                                for phase_col in phase_subset:
                                    phase_list = filt_data.get(
                                        phase_col, default=pd.Series()
                                    ).to_list()
                                    phase_list = [
                                        phase.replace(",", "").replace("N", "")
                                        for phase in phase_list
                                        if phase is not None
                                    ]
                                    phase_list = [
                                        list(phase) for phase in phase_list if len(phase) >= 1
                                    ]
                                    for ph_lst in phase_list:
                                        phases.extend(ph_lst)

            phases = set(phases)
            phases = phases.difference({None})
            phases = [phase for phase in list(phases) if len(phase) == 1]
            phases = sorted(phases)
            if not phases:
                phases = ["A", "B", "C"]
            all_phases.append(",".join(phases))

        bus_df["phase"] = all_phases
        return bus_df

    def get_system(self) -> DistributionSystem:
        return self.system



from ditto.readers.cim_iec_61968_13.components.distribution_bus import DistributionBusMapper
from ditto.readers.cim_iec_61968_13.components.distribution_capacitor import (
    DistributionCapacitorMapper,
)
from ditto.readers.cim_iec_61968_13.components.distribution_load import DistributionLoadMapper
from ditto.readers.cim_iec_61968_13.components.distribution_transformer import (
    DistributionTransformerMapper,
)
from ditto.readers.cim_iec_61968_13.equipment.distribution_transformer_equipment import (
    DistributionTransformerEquipmentMapper,
)
from ditto.readers.cim_iec_61968_13.equipment.matrix_branch_equipment import (
    MatrixImpedanceBranchEquipmentMapper,
)
from ditto.readers.cim_iec_61968_13.components.matrix_branch import MatrixImpedanceBranchMapper
from ditto.readers.cim_iec_61968_13.components.distribution_voltage_source import (
    DistributionVoltageSourceMapper,
)
from ditto.readers.cim_iec_61968_13.equipment.voltage_source_equipment import (
    VoltageSourceEquipmentMapper,
)
from ditto.readers.cim_iec_61968_13.controllers.regulator_controller import (
    RegulatorControllerMapper,
)
from ditto.readers.cim_iec_61968_13.components.distribution_regulator import (
    DistributionRegulatorMapper,
)

from ditto.readers.cim_iec_61968_13.components.matrix_impedance_switch import (
    MatrixImpedanceSwitchMapper,
)



from gdm.distribution.components import DistributionBus
from gdm.distribution.common import VoltageLimitSet
from gdm.distribution.enums import VoltageTypes
from gdm.quantities import Voltage
from infrasys.location import Location
import opendssdirect as odd
from loguru import logger

from ditto.readers.opendss.common import PHASE_MAPPER, get_equipment_from_catalog


def get_buses(crs: str = None) -> list[DistributionBus]:
    """Function to return list of all buses in Opendss model

    Args:
        crs (str, optional): Coordinate reference system name. Defaults to None.

    Returns:
        list[DistributionBus]: list of DistributionBus objects
    """

    logger.debug("parsing bus components...")
    voltage_limit_set_catalog = {}
    location_catalog = {}
    buses = []

    for bus in odd.Circuit.AllBusNames():
        odd.Circuit.SetActiveBus(bus)
        rated_voltage = odd.Bus.kVBase()

        loc = Location(x=odd.Bus.Y(), y=odd.Bus.X(), crs=crs)
        loc = get_equipment_from_catalog(loc, location_catalog)

        voltage_lower_bound = VoltageLimitSet(
            limit_type="min",
            value=Voltage(rated_voltage * 0.95, "kilovolt"),
        )
        voltage_lower_bound = get_equipment_from_catalog(
            voltage_lower_bound, voltage_limit_set_catalog
        )
        voltage_upper_bound = VoltageLimitSet(
            limit_type="max",
            value=Voltage(rated_voltage * 1.05, "kilovolt"),
        )
        voltage_upper_bound = get_equipment_from_catalog(
            voltage_upper_bound, voltage_limit_set_catalog
        )

        limitsets = [voltage_lower_bound, voltage_upper_bound]
        buses.append(
            DistributionBus.model_construct(
                voltage_type=VoltageTypes.LINE_TO_GROUND.value,
                name=bus,
                rated_voltage=Voltage(rated_voltage, "kilovolt"),
                phases=[PHASE_MAPPER[str(node)] for node in odd.Bus.Nodes()],
                coordinate=loc,
                voltagelimits=limitsets,
            )
        )
    return buses



from gdm.distribution.equipment import MatrixImpedanceBranchEquipment
from gdm.quantities import (
    CapacitancePULength,
    ResistancePULength,
    Current,
)
import numpy as np

from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper


class MatrixImpedanceBranchEquipmentMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        return MatrixImpedanceBranchEquipment(
            name=self.map_name(row),
            r_matrix=self.map_r_matrix(row),
            x_matrix=self.map_x_matrix(row),
            c_matrix=self.map_c_matrix(row),
            ampacity=self.map_ampacity(row),
        )

    def _array_to_matrix(self, array, n):
        array = [float(x) for x in array]
        matrix = np.zeros((n, n), dtype=float)
        index = 0
        for i in range(n):
            for j in range(i + 1):
                matrix[i, j] = array[index]
                matrix[j, i] = array[index]
                index += 1
        return matrix

    def map_name(self, row):
        return row["line_code"]

    def map_r_matrix(self, row):
        r = row["r"]
        r_matrix = self._array_to_matrix(r, int(row["phase_count"]))
        return ResistancePULength(r_matrix, "ohm/m")

    def map_x_matrix(self, row):
        x = row["x"]
        x_matrix = self._array_to_matrix(x, int(row["phase_count"]))
        return ResistancePULength(x_matrix, "ohm/m")

    def map_c_matrix(self, row):
        b = row["b"]
        b_matrix = self._array_to_matrix(b, int(row["phase_count"]))
        c_matrix = b_matrix / (2 * np.pi * 60)
        return CapacitancePULength(c_matrix, "F/m")

    def map_ampacity(self, row):
        return Current(row["ampacity_normal"], "ampere")



from gdm.distribution import DistributionSystem
from gdm.distribution.enums import Phase
from gdm.distribution.components import (
    DistributionTransformer,
    DistributionBranchBase,
    DistributionCapacitor,
    DistributionSolar,
    DistributionLoad,
)
from networkx import Graph, DiGraph

import networkx as nx

from ditto.readers.opendss.common import get_source_bus


def update_split_phase_nodes(graph: Graph, system: DistributionSystem) -> DistributionSystem:
    """Return the system with corrected split phase representation

    Args:
        graph (Graph):  Graph representation of the dirtribution model
        system (DistributionSystem): Instance of an gdm DistributionSystem

    Returns:
        DistributionSystem: Updated system with fixed split phase representation
    """

    source_buses = get_source_bus(system)
    assert len(set(source_buses)) == 1, "Source bus should be singular"
    tree = nx.dfs_tree(graph, source=source_buses[0])
    for u, v in tree.edges():
        attrs = graph.edges[u, v]
        tree.add_edge(u, v, **attrs)

    split_phase_transformers = _get_split_phase_transformers(system)
    for transformer in split_phase_transformers:
        _get_split_phase_sub_graph(transformer, tree, system)


def _get_split_phase_transformers(system: DistributionSystem) -> list[DistributionTransformer]:
    """returns a list of split phase transformers

    Args:
        system (DistributionSystem): Instance of an gdm DistributionSystem

    Returns:
        list[DistributionTransformer]: list of split phase transformers
    """

    split_phase_transformers = system.get_components(
        DistributionTransformer,
        filter_func=lambda x: x.equipment.is_center_tapped,
    )
    return list(split_phase_transformers)


def _get_split_phase_sub_graph(
    xfmr: DistributionTransformer, graph: DiGraph, system: DistributionSystem
) -> Graph:
    """returns the subgraph for a given distribution transformer

    Args:
        xfmr(DistributionTransformer): instance of DistributionTransformer
        graph (DiGraph): Graph representation of the dirtribution model
        system (DistributionSystem): Instance of an gdm DistributionSystem

    Returns:
        Graph: subgraph for a given distribution transformer
    """

    xfmr_buses = list({bus.name for bus in xfmr.buses})
    hv_xfmr_bus = (
        xfmr_buses[1]
        if xfmr_buses[0] in list(nx.dfs_successors(graph, xfmr_buses[1]))
        else xfmr_buses[0]
    )
    lv_xfmr_bus = (
        xfmr_buses[0]
        if xfmr_buses[0] in list(nx.dfs_successors(graph, xfmr_buses[1]))
        else xfmr_buses[1]
    )

    xfmr_info = graph[hv_xfmr_bus][lv_xfmr_bus]
    assert (
        xfmr_info["type"] == DistributionTransformer
    ), f"Unsupported model type {xfmr_info['type']}"
    model_type = xfmr_info["type"]
    xfmr_model = system.get_component(model_type, xfmr_info["name"])

    filter_nodes = []
    for node, sucessors in nx.bfs_successors(graph, lv_xfmr_bus):
        filter_nodes.append(node)
        filter_nodes.extend(sucessors)

    descendants = set(filter_nodes)
    subgraph = graph.subgraph(descendants)
    _get_components_in_subgraph(hv_xfmr_bus, xfmr_model, subgraph, system)
    return


def _get_components_in_subgraph(
    hv_xfmr_bus: str,
    xfmr_model: DistributionTransformer,
    subgraph: Graph,
    system: DistributionSystem,
):
    """fixes the split phase representation for the system

    Args:
        hv_xfmr_bus (str): bus name for the distribution transformer (HV side)
        xfmr_model (DistributionTransformer): instance of DistributionTransformer
        subgraph (Graph): subgraph (downstream) of a split phase dirtribution transformer
        system (DistributionSystem): Instance of an gdm DistributionSystem

    """

    secondary_wdg_phases = xfmr_model.winding_phases[1:]
    xfmr_phases_filtered = []
    split_phases = [Phase.S1, Phase.S2]
    xfmr_phases_filtered = [
        phase for phase_lists in secondary_wdg_phases for phase in phase_lists if phase != Phase.N
    ]
    mapped_split_phases = {k: v for k, v in zip(xfmr_phases_filtered, split_phases)}
    _fix_bus_phases(hv_xfmr_bus, mapped_split_phases, subgraph, system)
    _fix_transformer_phases(mapped_split_phases, xfmr_model, secondary_wdg_phases)
    _fix_component_phases(mapped_split_phases, subgraph, system)


def _fix_bus_phases(
    hv_xfmr_bus: str, mapped_split_phases: dict, subgraph: Graph, system: DistributionSystem
):
    """method fixes phasing for distribution buses downstream of split phase transformers

    Args:
        hv_xfmr_bus (str): bus name for the distribution transformer (HV side)
        mapped_split_phases (dict):  mapping A,B,C phase to S1/S2 phase
        subgraph (Graph): subgraph (downstream) of a split phase dirtribution transformer
        system (DistributionSystem): Instance of an gdm DistributionSystem
    """
    for _, _, data in subgraph.edges(data=True):
        model: DistributionBranchBase = system.get_component(data["type"], data["name"])
        assert issubclass(
            model.__class__, DistributionBranchBase
        ), f"Unsupported model type {model.__class__.__name__}"
        for bus in model.buses:
            if bus.name != hv_xfmr_bus:
                bus.phases = _mapped_phases(mapped_split_phases, bus.phases)
        model.phases = _mapped_phases(mapped_split_phases, model.phases)


def _mapped_phases(mapped_split_phases, phases):
    return list(
        set(
            [
                mapped_split_phases[phase] if phase in mapped_split_phases else phase
                for phase in phases
            ]
            + [Phase.N]
        )
    )


def _fix_transformer_phases(
    mapped_split_phases: dict,
    xfmr_model: DistributionTransformer,
    secondary_wdg_phases: list[list[Phase]],
):
    """method fixes phasing for split phase transformers

    Args:
        mapped_split_phases (dict): mapping A,B,C phase to S1/S2 phase
        xfmr_model (DistributionTransformer): instance of DistributionTransformer
        secondary_wdg_phases (list[list[Phase]]): secondary phases for a distribution transformers
    """
    wdg_phases = [xfmr_model.winding_phases[0]]
    for phase_list in secondary_wdg_phases:
        new_phases = list(
            set(
                [
                    mapped_split_phases[phase] if phase in mapped_split_phases else phase
                    for phase in phase_list
                ]
            )
        )
        wdg_phases.append(new_phases)
    xfmr_model.winding_phases = wdg_phases


def _fix_component_phases(mapped_split_phases: dict, subgraph: Graph, system: DistributionSystem):
    """method fixes phasing for components downstream of split phase transformers

    Args:
        mapped_split_phases (dict): mapping A,B,C phase to S1/S2 phase
        subgraph (Graph): subgraph (downstream) of a split phase dirtribution transformer
        system (DistributionSystem): Instance of an gdm DistributionSystem
    """
    component_types = [DistributionLoad, DistributionCapacitor, DistributionSolar]
    for node in subgraph.nodes():
        for component_type in component_types:
            for component in system.get_bus_connected_components(node, component_type) or []:
                component.phases = [mapped_split_phases[phase] for phase in component.phases]



from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class DistributionTransformerEquipmentMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "XfmrCode_X12X13X23"
    altdss_composition_name = "XfmrCode"
    opendss_file = OpenDSSFileTypes.TRANSFORMERS_FILE.value

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_pct_no_load_loss(self):
        self.opendss_dict["pctNoLoadLoss"] = self.model.pct_no_load_loss

    def map_pct_full_load_loss(self):
        self.opendss_dict["pctLoadLoss"] = self.model.pct_full_load_loss

    def map_windings(self):
        kvs = []
        pctRs = []
        kVAs = []
        conns = []
        taps = []
        min_tap = []
        max_tap = []
        num_taps = []
        # TODO: Add TapWindingEquipment
        for i in range(len(self.model.windings)):
            winding = self.model.windings[i]
            tap_pu = winding.tap_positions
            taps.append(tap_pu[0])
            min_tap.append(winding.min_tap_pu)
            max_tap.append(winding.max_tap_pu)
            num_taps.append(winding.total_taps)

            num_phases = winding.num_phases
            # rated_voltage
            nom_voltage = winding.rated_voltage.to("kV").magnitude
            kvs.append(nom_voltage if num_phases == 1 else nom_voltage * 1.732)
            # resistance
            pctRs.append(winding.resistance)
            # rated_power
            kVAs.append(winding.rated_power.to("kva").magnitude)
            # connection_type
            conns.append(self.connection_map[winding.connection_type])
            # TODO: num_phases and is_grounded aren't included
            if self.model.is_center_tapped and i == len(self.model.windings):
                kvs.append(nom_voltage)
                pctRs.append(winding.resistance)
                kVAs.append(winding.rated_power.to("kVa").magnitude)
                conns.append(self.connection_map[winding.connection_type])
        self.opendss_dict["kV"] = kvs
        self.opendss_dict["pctR"] = pctRs
        self.opendss_dict["kVA"] = kVAs
        self.opendss_dict["Conn"] = conns
        for x, x_value in zip(["X12", "X23", "X13"], self.model.winding_reactances):
            self.opendss_dict[x] = x_value
        self.opendss_dict["Phases"] = num_phases
        self.opendss_dict["Tap"] = taps
        self.opendss_dict["Tap"] = taps
        self.opendss_dict["MinTap"] = min_tap
        self.opendss_dict["MaxTap"] = max_tap
        self.opendss_dict["NumTaps"] = num_taps
        pass

    def map_coupling_sequences(self):
        # Used to know the reactance couplings
        pass

    def map_winding_reactances(self):
        for i in range(len(self.model.coupling_sequences)):
            coupling_sequence = self.model.coupling_sequences[i]
            reactance = self.model.winding_reactances[i]
            first = coupling_sequence[0] + 1
            second = coupling_sequence[1] + 1
            reactance_name = "X" + str(first) + str(second)
            self.opendss_dict[reactance_name] = reactance

    def map_is_center_tapped(self):
        pass  # Used on buses

    def map_mounting(self):
        pass



from gdm.distribution.components import MatrixImpedanceBranch, DistributionBus
from gdm.distribution.equipment import MatrixImpedanceBranchEquipment
from gdm.quantities import Distance

from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper
from ditto.readers.cim_iec_61968_13.common import phase_mapper


class MatrixImpedanceBranchMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        return MatrixImpedanceBranch(
            name=self.map_name(row),
            buses=self.map_buses(row),
            length=self.map_length(row),
            phases=self.map_phases(row),
            equipment=self.map_equipment(row),
        )

    def map_name(self, row):
        return row["line"]

    def map_buses(self, row):
        bus_1_name = row["bus_1"]
        bus_1 = self.system.get_component(DistributionBus, bus_1_name)
        bus_2_name = row["bus_2"]
        bus_2 = self.system.get_component(DistributionBus, bus_2_name)
        return [bus_1, bus_2]

    def map_length(self, row):
        length = float(row["length"])
        return Distance(length, "m")

    def map_phases(self, row):
        phases = row["phases_1"].split(",")
        return [phase_mapper[phase] for phase in phases]

    def map_equipment(self, row):
        equipment = self.system.get_component(MatrixImpedanceBranchEquipment, row["line_code"])
        return equipment



from pathlib import Path

from gdm.distribution.common import SequencePair
from gdm.distribution import DistributionSystem

from pydantic import ValidationError
from rich.console import Console
from infrasys import Component
from rich.table import Table
import opendssdirect as odd
from loguru import logger

from ditto.readers.opendss.components.conductors import get_conductors_equipment
from ditto.readers.opendss.components.cables import get_cables_equipment
from ditto.readers.opendss.components.sources import get_voltage_sources
from ditto.readers.opendss.components.capacitors import get_capacitors
from ditto.readers.opendss.graph_utils import update_split_phase_nodes
from ditto.readers.opendss.components.pv_systems import get_pvsystems
from ditto.readers.opendss.components.buses import get_buses
from ditto.readers.opendss.components.loads import get_loads
from ditto.readers.opendss.components.transformers import (
    get_transformer_equipments,
    get_transformers,
)
from ditto.readers.opendss.components.branches import (
    get_geometry_branch_equipments,
    get_matrix_branch_equipments,
    get_branches,
)

from ditto.readers.reader import AbstractReader


SEQUENCE_PAIRS = [SequencePair(1, 2), SequencePair(1, 3), SequencePair(2, 3)]


class Reader(AbstractReader):
    """Class interface for Opendss case file reader"""

    validation_errors = []

    def __init__(self, Opendss_master_file: Path, crs: str | None = None) -> None:
        """Constructor for the Opendss reader

        Args:
            Opendss_master_file (Path): Path to the Opendss master file
            crs (str | None, optional): Coordinate reference system name. Defaults to None.
        """

        self.system = DistributionSystem(auto_add_composed_components=True)
        self.Opendss_master_file = Path(Opendss_master_file)
        self.crs = crs
        self._read()

    def _add_components(self, components: list[Component]):
        """Internal method to add components to the system."""

        if components:
            for component in components:
                try:
                    component.__class__.model_validate(component.model_dump())
                except ValidationError as e:
                    for error in e.errors():
                        self.validation_errors.append(
                            [
                                component.name,
                                component.__class__.__name__,
                                error["loc"][0] if error["loc"] else "On model validation",
                                error["type"],
                                error["msg"],
                            ]
                        )

            self.system.add_components(*components)

    def _read(self):
        """Takes the master file path and returns instance of OpendssParser

        Raises:
            FileNotFoundError: Error raised if the file is not found
        """

        logger.debug("Loading OpenDSS model.")
        if not self.Opendss_master_file.exists():
            msg = f"File not found: {self.Opendss_master_file}"
            raise FileNotFoundError(msg)

        odd.Text.Command("Clear")
        odd.Basic.ClearAll()
        odd.Text.Command(f'Redirect "{self.Opendss_master_file}"')
        odd.Text.Command("Solve")
        logger.debug(f"Model loaded from {self.Opendss_master_file}.")

        odd.Solution.Solve()

        self._add_components(get_buses(self.crs))
        self._add_components(get_voltage_sources(self.system))
        self._add_components(get_capacitors(self.system))
        self._add_components(get_loads(self.system))
        self._add_components(get_pvsystems(self.system))
        (
            distribution_transformer_equipment_catalog,
            winding_equipment_catalog,
        ) = get_transformer_equipments(self.system)
        self._add_components(distribution_transformer_equipment_catalog.values())
        self._add_components(
            get_transformers(
                self.system, distribution_transformer_equipment_catalog, winding_equipment_catalog
            )
        )
        self._add_components(get_conductors_equipment())
        self._add_components(get_cables_equipment())
        matrix_branch_equipments_catalog, thermal_limit_catalog = get_matrix_branch_equipments()
        for catalog in matrix_branch_equipments_catalog:
            self._add_components(matrix_branch_equipments_catalog[catalog].values())

        geometry_branch_equipment_catalog, mapped_geometry = get_geometry_branch_equipments(
            self.system
        )
        self._add_components(geometry_branch_equipment_catalog.values())
        branches = get_branches(
            self.system,
            mapped_geometry,
            geometry_branch_equipment_catalog,
            matrix_branch_equipments_catalog,
            thermal_limit_catalog,
        )
        self._add_components(branches)

        logger.debug("parsing complete...")
        logger.debug(f"\n{self.system.info()}")
        logger.debug("Building graph...")
        graph = self.system.get_undirected_graph()
        logger.debug(graph)
        logger.debug("Graph build complete...")
        logger.debug("Updating graph to fix split phase representation...")
        update_split_phase_nodes(graph, self.system)
        logger.debug("System update complete...")
        self._validate_model()

    def get_system(self) -> DistributionSystem:
        """Returns an instance of DistributionSystem

        Returns:
            DistributionSystem: Instance of DistributionSystem
        """

        return self.system

    def _validate_model(self):
        if self.validation_errors:
            error_table = Table(title="Validation warning summary")
            error_table.add_column("Model", justify="right", style="cyan", no_wrap=True)
            error_table.add_column("Type", style="green")
            error_table.add_column("Field", justify="right", style="bright_magenta")
            error_table.add_column("Error", style="bright_red")
            error_table.add_column("Message", justify="right", style="turquoise2")

            for row in self.validation_errors:
                print(row)
                error_table.add_row(*row)

            console = Console()
            console.print(error_table)
            raise Exception(
                "Validations errors occured when running the script. See the table above"
            )

        ...



from uuid import uuid4
from enum import Enum

from infrasys import System, Component
from infrasys.quantities import Time

from gdm.quantities import (
    CapacitancePULength,
    ResistancePULength,
    ReactancePULength,
    Distance,
    Current,
)

from gdm.distribution.common import TimeCurrentCurve, ThermalLimitSet
from gdm.distribution.equipment import (
    MatrixImpedanceRecloserEquipment,
    MatrixImpedanceSwitchEquipment,
    MatrixImpedanceBranchEquipment,
    MatrixImpedanceFuseEquipment,
    ConcentricCableEquipment,
    GeometryBranchEquipment,
    BareConductorEquipment,
)
from gdm.distribution.components import (
    MatrixImpedanceSwitch,
    MatrixImpedanceBranch,
    MatrixImpedanceFuse,
    DistributionBus,
    GeometryBranch,
)
from gdm.distribution.enums import Phase

import opendssdirect as odd
from loguru import logger
import numpy as np

from ditto.readers.opendss.common import (
    get_equipment_from_catalog,
    query_model_data,
    PHASE_MAPPER,
    UNIT_MAPPER,
    hash_model,
)


class MatrixBranchTypes(str, Enum):
    LINE_CODE = "LineCodes"
    LINE = "Lines"


def get_geometry_branch_equipments(
    system: System,
) -> tuple[list[GeometryBranchEquipment], dict[str, int]]:
    """Helper function that return a list of GeometryBranchEquipment objects

    Args:
        system (System): Instance of System

    Returns:
        list[GeometryBranchEquipment]: list of GeometryBranchEquipment objects
        dict[str, int]: mapping of line geometries names to GeometryBranchEquipment hash
    """

    logger.debug("parsing geometry branch equipment...")

    mapped_geometry = {}
    geometry_branch_equipment_catalog = {}
    flag = odd.LineGeometries.First()

    while flag > 0:
        geometry_name = odd.LineGeometries.Name().lower()
        x_coordinates = []
        y_coordinates = []
        units = UNIT_MAPPER[odd.LineGeometries.Units()[0].value]
        odd.Text.Command(f"? LineGeometry.{geometry_name}.wires")
        wires = odd.Text.Result().strip("[]").split(", ")
        model_name = odd.Element.Name().lower().split(".")[1]
        conductor_elements = []

        for i, wire in enumerate(wires):
            odd.Text.Command(f"LineGeometry.{geometry_name}.cond={i + 1}")
            odd.Text.Command(f"? LineGeometry.{geometry_name}.h")
            y_coordinates.append(float(odd.Text.Result()))
            odd.Text.Command(f"? LineGeometry.{geometry_name}.x")
            x_coordinates.append(float(odd.Text.Result()))

            equipments = list(
                system.get_components(BareConductorEquipment, filter_func=lambda x: x.name == wire)
            )
            if not equipments:
                equipments = list(
                    system.get_components(
                        ConcentricCableEquipment, filter_func=lambda x: x.name == wire
                    )
                )
            equipment = equipments[0]
            conductor_elements.append(equipment)

        geometry_branch_equipment = GeometryBranchEquipment.model_construct(
            name=model_name,
            conductors=conductor_elements,
            horizontal_positions=Distance(x_coordinates, units),
            vertical_positions=Distance(y_coordinates, units),
        )
        geometry_branch_equipment = get_equipment_from_catalog(
            geometry_branch_equipment, geometry_branch_equipment_catalog
        )
        mapped_geometry[geometry_name] = hash_model(geometry_branch_equipment)

        flag = odd.LineGeometries.Next()

    return geometry_branch_equipment_catalog, mapped_geometry


def _build_matrix_branch(
    model_type: str,
    matrix_branch_equipments_catalog: dict[int, Component],
    thermal_limit_catalog: dict[int, Component],
    model_class: (type[MatrixImpedanceSwitchEquipment] | type[MatrixImpedanceBranchEquipment]),
    fuse: dict,
    recloser: dict,
) -> MatrixImpedanceBranchEquipment:
    """Helper function to build a MatrixImpedanceBranchEquipment instance

    Args:
        model_type (str): OpenDSS model type e.g. LinesCode / Line
        matrix_branch_equipments_catalog (dict[int, Component]): mapping of model hash to MatrixImpedanceBranchEquipment instance
        thermal_limit_catalog (dict[int, Component]): mapping of model hash to ThermalLimitSet instance

    Returns:
        MatrixImpedanceBranchEquipment: instance of MatrixImpedanceBranchEquipment
    """

    model_name = odd.Element.Name().lower().split(".")[1]
    if model_type == MatrixBranchTypes.LINE_CODE.value:
        equipment_uuid = model_name
    else:
        equipment_uuid = str(uuid4())
    module: odd.LineCodes | odd.Lines = getattr(odd, model_type)

    num_phase = module.Phases()
    length_units = UNIT_MAPPER[module.Units().value]

    r_matrix = module.RMatrix() if model_type == MatrixBranchTypes.LINE.value else module.Rmatrix()
    x_matrix = module.XMatrix() if model_type == MatrixBranchTypes.LINE.value else module.Xmatrix()
    c_matrix = module.CMatrix() if model_type == MatrixBranchTypes.LINE.value else module.Cmatrix()
    amps = module.NormAmps() if module.NormAmps() else 0.001
    matrix_branch_dict = {
        "name": equipment_uuid,
        "r_matrix": ResistancePULength(
            np.reshape(np.array(r_matrix), (num_phase, num_phase)),
            f"ohm/{length_units}",
        ),
        "x_matrix": ReactancePULength(
            np.reshape(np.array(x_matrix), (num_phase, num_phase)),
            f"ohm/{length_units}",
        ),
        "c_matrix": CapacitancePULength(
            np.reshape(np.array(c_matrix), (num_phase, num_phase)),
            f"nanofarad/{length_units}",
        ),
        "ampacity": Current(amps, "ampere"),
    }
    if model_class == MatrixImpedanceSwitchEquipment:
        # TODO: implement switch controller logic here
        controller = _build_distribution_switch_controller()
        matrix_branch_dict["controller"] = controller
    elif model_class == MatrixImpedanceFuseEquipment:
        matrix_branch_dict.update(fuse)
    elif model_class == MatrixImpedanceRecloserEquipment:
        matrix_branch_dict.update(recloser)
    matrix_branch_equipment = model_class.model_construct(**matrix_branch_dict)
    matrix_branch_equipment = get_equipment_from_catalog(
        matrix_branch_equipment, matrix_branch_equipments_catalog, model_class.__name__
    )
    return matrix_branch_equipment


def _build_distribution_switch_controller():
    return None


def get_matrix_branch_equipments() -> (
    tuple[dict[int, MatrixImpedanceBranchEquipment], dict[int, ThermalLimitSet]]
):
    """Function to return list of all MatrixImpedanceBranchEquipment in Opendss model.

    Returns:
        dict[int, MatrixImpedanceBranchEquipment]: mapping of model hash to MatrixImpedanceBranchEquipment instance
        dict[int, ThermalLimitSet]: mapping of model hash to ThermalLimitSet instance
    """

    logger.debug("parsing matrix branch equipment...")
    reclosers = get_reclosers()
    fuses = get_fuses()
    matrix_branch_equipments_catalog = {
        MatrixImpedanceRecloserEquipment.__name__: {},
        MatrixImpedanceSwitchEquipment.__name__: {},
        MatrixImpedanceBranchEquipment.__name__: {},
        MatrixImpedanceFuseEquipment.__name__: {},
    }
    thermal_limit_catalog = {}
    odd_model_types = [v.value for v in MatrixBranchTypes]
    for odd_model_type in odd_model_types:
        module: odd.LineCodes | odd.Lines = getattr(odd, odd_model_type)
        flag = module.First()
        while flag > 0:
            if odd_model_type == MatrixBranchTypes.LINE.value and module.Geometry():
                pass
            else:
                fuse = {}
                recloser = {}
                if (
                    odd_model_type == MatrixBranchTypes.LINE.value
                    and odd.Lines.Name().lower() in fuses
                ):
                    fuse = fuses[odd.Lines.Name().lower()]
                    model_type = MatrixImpedanceFuseEquipment
                elif (
                    odd_model_type == MatrixBranchTypes.LINE.value
                    and odd.Lines.Name().lower() in reclosers
                ):
                    recloser = reclosers[odd.Lines.Name().lower()]
                    model_type = MatrixImpedanceRecloserEquipment
                elif odd_model_type == MatrixBranchTypes.LINE.value and odd.Lines.IsSwitch():
                    model_type = MatrixImpedanceSwitchEquipment
                else:
                    model_type = MatrixImpedanceBranchEquipment
                _build_matrix_branch(
                    odd_model_type,
                    matrix_branch_equipments_catalog,
                    thermal_limit_catalog,
                    model_type,
                    fuse,
                    recloser,
                )
            flag = module.Next()
    return matrix_branch_equipments_catalog, thermal_limit_catalog


def get_branches(
    system: System,
    mapping: dict[str, str],
    geometry_branch_equipment_catalog: dict,
    matrix_branch_equipments_catalog: dict,
    thermal_limit_catalog: dict,
) -> tuple[list[MatrixImpedanceBranch | GeometryBranch]]:
    """Method to build a model branches

    Args:
        system (System): Instance of System
        mapping (dict[str, int]): mapping of line geometries names to GeometryBranchEquipment hash
        geometry_branch_equipment_catalog (dict): mapping of model hash to GeometryBranchEquipment instance
        matrix_branch_equipments_catalog (dict): mapping of model hash to MatrixImpedanceBranchEquipment instance
        thermal_limit_catalog (dict): mapping of model hash to ThermalLimitSet instance

    Returns:
        tuple[list[MatrixImpedanceBranch | GeometryBranch]]: Returns a list of system branches
    """

    logger.debug("parsing branch components...")
    reclosers = get_reclosers()
    fuses = get_fuses()
    branches = []
    flag = odd.Lines.First()
    while flag > 0:
        logger.debug(f"building line {odd.CktElement.Name()}...")

        buses = odd.CktElement.BusNames()
        bus1, bus2 = buses[0].split(".")[0], buses[1].split(".")[0]
        num_phase = odd.CktElement.NumPhases()
        nodes = ["1", "2", "3"] if num_phase == 3 else buses[0].split(".")[1:]
        geometry = odd.Lines.Geometry().lower()
        if geometry:
            assert geometry in mapping
            geometry_hash = mapping[geometry]
            geometry_branch_equipment = geometry_branch_equipment_catalog[geometry_hash]

            n_conds = len(geometry_branch_equipment.conductors)
            for _ in range(n_conds - num_phase):
                nodes.append("4")

            if "4" in nodes:
                for bus in [bus1, bus2]:
                    bus_obj = system.get_component(DistributionBus, bus)
                    bus_obj.phases.append(Phase.N)

            geometry_branch = GeometryBranch.model_construct(
                name=odd.Lines.Name().lower(),
                equipment=geometry_branch_equipment,
                buses=[
                    system.get_component(DistributionBus, bus1),
                    system.get_component(DistributionBus, bus2),
                ],
                length=Distance(odd.Lines.Length(), UNIT_MAPPER[odd.Lines.Units()]),
                phases=[PHASE_MAPPER[node] for node in nodes],
            )
            branches.append(geometry_branch)
        else:
            fuse = {}
            recloser = {}
            if odd.Lines.Name().lower() in fuses:
                fuse = fuses[odd.Lines.Name().lower()]
                equipment_class = MatrixImpedanceFuseEquipment
                model_class = MatrixImpedanceFuse
            elif odd.Lines.Name().lower() in reclosers:
                recloser = reclosers[odd.Lines.Name().lower()]
                equipment_class = MatrixImpedanceFuseEquipment
                model_class = MatrixImpedanceFuse
            elif odd.Lines.IsSwitch():
                equipment_class = MatrixImpedanceSwitchEquipment
                model_class = MatrixImpedanceSwitch
            else:
                equipment_class = MatrixImpedanceBranchEquipment
                model_class = MatrixImpedanceBranch
            equipment = _build_matrix_branch(
                MatrixBranchTypes.LINE.value,
                matrix_branch_equipments_catalog,
                thermal_limit_catalog,
                equipment_class,
                fuse,
                recloser,
            )
            equipment = get_equipment_from_catalog(
                equipment, matrix_branch_equipments_catalog, equipment_class.__name__
            )
            model_dict = {
                "name": odd.Lines.Name().lower(),
                "buses": [
                    system.get_component(DistributionBus, bus1),
                    system.get_component(DistributionBus, bus2),
                ],
                "length": Distance(odd.Lines.Length(), UNIT_MAPPER[odd.Lines.Units()]),
                "phases": [PHASE_MAPPER[node] for node in nodes],
                "equipment": equipment,
            }
            if model_class in [MatrixImpedanceSwitch, MatrixImpedanceFuse]:
                model_dict["is_closed"] = [
                    not (odd.CktElement.IsOpen(1, node + 1) or odd.CktElement.IsOpen(2, node + 1))
                    for node in range(len(nodes))
                ]
            matrix_branch = model_class.model_construct(**model_dict)
            branches.append(matrix_branch)
        flag = odd.Lines.Next()

    return branches


def get_tcc_curves() -> dict[str, TimeCurrentCurve]:
    """method returns a dict of tcc curve names mapped to TimeCurrentCurve objects

    Returns:
        dict[str, TimeCurrentCurve]: mapped TimeCurrentCurve objects
    """
    curves = {}
    odd.Circuit.SetActiveClass("tcc_curve")
    flag = odd.ActiveClass.First()
    while flag:
        element_type, element_name = odd.Element.Name().split(".")
        c_array = query_model_data(element_type, element_name, "c_array", list)
        t_array = query_model_data(element_type, element_name, "t_array", list)
        curves[element_name] = TimeCurrentCurve(
            curve_x=Current(c_array, "ampere"),
            curve_y=Time(t_array, "second"),
        )
        flag = odd.ActiveClass.Next()
    return curves


def get_fuses() -> list[str]:
    """Returns a list of lines with fuses

    Returns:
        list[str]: List of lines with fuses
    """
    curves = get_tcc_curves()
    lines_with_fuses = {}
    flag = odd.Fuses.First()
    while flag:
        _, object_name = odd.Fuses.MonitoredObj().split(".")

        curve_name = odd.Fuses.TCCCurve()
        lines_with_fuses[object_name] = {
            "delay": Time(odd.Fuses.Delay(), "seconds"),
            "tcc_curve": curves[curve_name],
        }
        flag = odd.Fuses.Next()
    return lines_with_fuses


def get_reclosers():
    """Returns a list of lines with reclosers

    Returns:
        list[str]: List of lines with reclosers
    """
    lines_with_reclosers = {}
    flag = odd.Reclosers.First()
    while flag:
        _, object_name = odd.Reclosers.MonitoredObj().split(".")
        lines_with_reclosers[object_name] = None
        flag = odd.Reclosers.Next()
    return lines_with_reclosers



from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class MatrixImpedanceBranchEquipmentMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "LineCode_ZMatrixCMatrix"
    altdss_composition_name = "LineCode"
    opendss_file = OpenDSSFileTypes.LINECODES_FILE.value

    def map_common(self):
        r_matrix_ohms = self.model.r_matrix.to("ohm/km")
        x_matrix_ohms = self.model.x_matrix.to("ohm/km")
        c_matrix_ohms = self.model.c_matrix.to("nanofarad/km")
        assert len(r_matrix_ohms) == len(x_matrix_ohms) == len(c_matrix_ohms)
        self.opendss_dict["NPhases"] = len(r_matrix_ohms.magnitude)
        self.opendss_dict["Units"] = "km"

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_r_matrix(self):
        r_matrix_ohms = self.model.r_matrix.to("ohm/km")
        self.opendss_dict["RMatrix"] = r_matrix_ohms.magnitude

    def map_x_matrix(self):
        x_matrix_ohms = self.model.x_matrix.to("ohm/km")
        self.opendss_dict["XMatrix"] = x_matrix_ohms.magnitude

    def map_c_matrix(self):
        c_matrix_nf = self.model.c_matrix.to("nanofarad/km")
        self.opendss_dict["CMatrix"] = c_matrix_nf.magnitude

    def map_ampacity(self):
        ampacity_amps = self.model.ampacity.to("ampere")
        self.opendss_dict["NormAmps"] = ampacity_amps.magnitude

    def map_construction(self):
        pass



from abc import ABC, abstractmethod

from gdm.distribution import DistributionSystem


class AbstractWriter(ABC):
    def __init__(self, system: DistributionSystem):
        self.system = system

    @abstractmethod
    def write(self) -> None:
        ...



from ditto.writers.opendss.components.distribution_branch import DistributionBranchMapper
from ditto.enumerations import OpenDSSFileTypes


class GeometryBranchMapper(DistributionBranchMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "Line_LineGeometry"
    altdss_composition_name = "Line"
    opendss_file = OpenDSSFileTypes.LINES_FILE.value

    def map_equipment(self):
        self.opendss_dict["Geometry"] = self.model.equipment.name

    def map_in_service(self):
        self.opendss_dict["enabled"] = self.model.in_service



from math import pi

from gdm.distribution.equipment import VoltageSourceEquipment, PhaseVoltageSourceEquipment
from gdm.quantities import Resistance, Reactance, Angle, Voltage
from gdm.distribution.enums import VoltageTypes

from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper


class VoltageSourceEquipmentMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        return VoltageSourceEquipment(
            name=self.map_name(row),
            sources=self.map_sources(row),
        )

    def map_name(self, row):
        return row["source"] + "_equipment"

    def map_sources(self, row):
        return [
            PhaseVoltageSourceEquipmentMapper(self.system).parse(row, phase)
            for phase in ["A", "B", "C"]
        ]


class PhaseVoltageSourceEquipmentMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row, phase):
        return PhaseVoltageSourceEquipment(
            name=self.map_name(row, phase),
            r0=self.map_r0(row),
            r1=self.map_r1(row),
            x0=self.map_x0(row),
            x1=self.map_x1(row),
            voltage=self.map_voltage(row),
            voltage_type=VoltageTypes.LINE_TO_GROUND,
            angle=self.map_angle(row),
        )

    def map_name(self, row, phase):
        return row["source"] + "_phase_voltage_source_equipment_" + phase

    def map_r0(self, row):
        return Resistance(float(row["r0"]), "ohm")

    def map_r1(self, row):
        return Resistance(float(row["r1"]), "ohm")

    def map_x0(self, row):
        return Reactance(float(row["x0"]), "ohm")

    def map_x1(self, row):
        return Reactance(float(row["x1"]), "ohm")

    def map_voltage(self, row):
        return Voltage(float(row["src_voltage"]) / 1.732, "volt")

    def map_angle(self, row):
        return Angle(float(row["src_angle"]) * 180 * pi, "degree")



from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class DistributionBusMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "Bus"
    altdss_composition_name = None
    opendss_file = OpenDSSFileTypes.COORDINATE_FILE.value

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_coordinate(self):
        if hasattr(self.model.coordinate, "x"):
            self.opendss_dict["X"] = self.model.coordinate.y
        if hasattr(self.model.coordinate, "y"):
            self.opendss_dict["Y"] = self.model.coordinate.x

    def map_rated_voltage(self):
        kv_rated_voltage = self.model.rated_voltage.to("kV")
        if self.model.voltage_type == "line-to-ground":
            self.opendss_dict["kVLN"] = kv_rated_voltage.magnitude
        elif self.model.voltage_type == "line-to-line":
            self.opendss_dict["kVLL"] = kv_rated_voltage.magnitude

    def map_phases(self):
        # Not mapped for OpenDSS buses in BusCoords.dss files
        return

    def map_voltagelimits(self):
        # Not mapped for OpenDSS buses in BusCoords.dss files
        return

    def map_voltage_type(self):
        # Handled in map_rated_voltage
        return

    def map_in_service(self):
        return



from datetime import datetime, timedelta
from enum import Enum
from gdm.distribution.components import (
    DistributionVoltageSource,
    DistributionSolar,
    DistributionLoad,
)
from infrasys.normalization import NormalizationMax, NormalizationByValue
from gdm.quantities import ActivePower, ReactivePower, Irradiance
from infrasys.time_series_models import SingleTimeSeries
from infrasys.base_quantity import BaseQuantity
from pydantic import BaseModel
import opendssdirect as odd
from loguru import logger


class ObjectsWithProfile(Enum):
    """GDM models that support profiles in OpenDSS"""

    LOAD = DistributionLoad
    PV_SYSTEM = DistributionSolar
    SOURCE = DistributionVoltageSource


class ProfileBases(str, Enum):
    """Profile base (used for normalization of profiles)"""

    P_BASE = "PBase"
    Q_BASE = "QBase"


class ProfileTypes(str, Enum):
    """Profile multiplier (used for scaling profiles)"""

    P_MULT = "PMult"
    Q_MULT = "QMult"


class ProfileMap(BaseModel):
    """Profile mapping"""

    quantity: type[BaseQuantity] | None
    units: str | None
    profile_type: ProfileTypes
    variable: str

    class Config:
        arbitrary_types_allowed = True


profile_type_to_base_type_map = {
    ProfileTypes.P_MULT: ProfileBases.P_BASE,
    ProfileTypes.Q_MULT: ProfileBases.Q_BASE,
}

model_type_to_profile_type_map = {
    ObjectsWithProfile.LOAD: [
        ProfileMap(
            profile_type=ProfileTypes.P_MULT,
            quantity=ActivePower,
            units="kilowatt",
            variable="active_power",
        ),
        ProfileMap(
            profile_type=ProfileTypes.Q_MULT,
            quantity=ReactivePower,
            units="kilovar",
            variable="reactive_power",
        ),
    ],
    ObjectsWithProfile.PV_SYSTEM: [
        ProfileMap(
            profile_type=ProfileTypes.P_MULT,
            quantity=Irradiance,
            units="kilowatt/meter**2",
            variable="irradiance",
        ),
    ],
    ObjectsWithProfile.SOURCE: [
        ProfileMap(
            profile_type=ProfileTypes.P_MULT, quantity=None, units=None, variable="pu_voltage"
        ),
    ],
}


def build_profiles(
    profile_names: list[str], component_type: ObjectsWithProfile, profile_catalog: dict[str,]
) -> dict[str, dict[str, SingleTimeSeries]]:
    """Function to return dictionary of SingleTimeSeries objects representing load shapes in the Opendss model.

    Args:
        profile_names (list[str]): list of profile names
        component_type (ObjectsWithProfile): type of component
        profile_catalog (dict[str,]): dictionary name mapping to  SingleTimeSeries objects in the
                                      following convention:
                                      dict[profile_name, dict[profile_type, SingleTimeSeries]]

    Returns:
        dict[str, dict[str, SingleTimeSeries]]: updated profile catalog is returned
    """

    logger.debug("parsing timeseries components...")
    for profile_name in profile_names:
        if profile_name and profile_name not in profile_catalog:
            profiles = {}
            odd.LoadShape.Name(profile_name)
            for profile in model_type_to_profile_type_map[component_type]:
                profile_type = profile.profile_type
                profile_base = profile_type_to_base_type_map[profile_type]
                profile_func = getattr(odd.LoadShape, profile_type.value)
                base_func = getattr(odd.LoadShape, profile_base.value)
                data = profile_func()
                if len(data) > 1:
                    if profile.quantity:
                        data = profile.quantity(data, profile.units)
                    length = odd.LoadShape.Npts()
                    initial_time = datetime(year=2020, month=1, day=1)
                    resolution = odd.LoadShape.SInterval()
                    time_array = [
                        initial_time + timedelta(seconds=i * resolution) for i in range(length)
                    ]
                    variable_name = profile.variable

                    if odd.LoadShape.Normalize() and base_func():
                        normalization = NormalizationByValue(value=base_func())
                    elif odd.LoadShape.Normalize() and not base_func():
                        normalization = NormalizationMax()
                    else:
                        normalization = None

                    ts = SingleTimeSeries.from_time_array(
                        data, variable_name, time_array, normalization=normalization
                    )
                    profiles[profile_type.value] = {
                        "data": ts,
                        "use_actual": odd.LoadShape.UseActual(),
                    }
            profile_catalog[profile_name] = profiles

    return profile_catalog



from gdm.quantities import Current, Distance, ResistancePULength
from gdm.distribution.equipment import BareConductorEquipment
import opendssdirect as odd
from loguru import logger

from ditto.readers.opendss.common import query_model_data, get_equipment_from_catalog


def get_conductors_equipment() -> list[BareConductorEquipment]:
    """Method returns a list of BareConductorEquipment objects

    Returns:
        list[BareConductorEquipment]: list of BareConductorEquipment
    """

    logger.debug("parsing conductor components...")
    bare_conductor_equipment_catalog = {}
    model_type = "WireData"
    conductors = []
    odd.Circuit.SetActiveClass(model_type)
    flag = odd.ActiveClass.First()
    while flag > 0:
        model_name = odd.Element.Name().lower().split(".")[1]
        gmr_units = query_model_data(model_type, model_name, "gmrunits", str)
        radius_units = query_model_data(model_type, model_name, "radunits", str)
        length_units = query_model_data(model_type, model_name, "runits", str)
        gmr = query_model_data(model_type, model_name, "gmr", float)
        diam = query_model_data(model_type, model_name, "diam", float)
        conductor = BareConductorEquipment.model_construct(
            emergency_ampacity=Current(
                query_model_data(model_type, model_name, "emergamps", float), "ampere"
            ),
            conductor_diameter=Distance(diam if diam else gmr / 0.7788, f"{radius_units}"),
            conductor_gmr=Distance(gmr if gmr else diam * 0.7788, f"{gmr_units}"),
            ac_resistance=ResistancePULength(
                query_model_data(model_type, model_name, "rac", float), f"ohms/{length_units}"
            ),
            dc_resistance=ResistancePULength(
                query_model_data(model_type, model_name, "rdc", float), f"ohms/{length_units}"
            ),
            ampacity=Current(
                query_model_data(model_type, model_name, "normamps", float), "ampere"
            ),
            name=model_name,
        )

        conductor = get_equipment_from_catalog(conductor, bare_conductor_equipment_catalog)
        conductors.append(conductor)
        flag = odd.ActiveClass.Next()
    return conductors



from gdm.distribution.controllers import RegulatorController
from gdm.distribution.components import DistributionBus
from gdm.quantities import Voltage
from infrasys.quantities import Time

from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper
from ditto.readers.cim_iec_61968_13.common import phase_mapper


class RegulatorControllerMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        return RegulatorController(
            name=self.map_name(row),
            delay=self.map_delay(row),
            v_setpoint=self.map_vsetpoint(row),
            min_v_limit=self.min_v_limit(row),
            max_v_limit=self.max_v_limit(row),
            pt_ratio=self.map_pt_ratio(row),
            use_ldc=self.map_use_ldc(row),
            is_reversible=self.map_is_reversible(row),
            ldc_R=self.map_ldc_R(row),
            ldc_X=self.map_ldc_X(row),
            ct_primary=self.map_ct_primary(row),
            max_step=self.map_max_step(row),
            bandwidth=self.map_bandwidth(row),
            controlled_bus=self.map_controlled_bus(row),
            controlled_phase=self.map_controlled_phase(row),
        )

    def map_name(self, row):
        return row["regulator"]

    def map_use_ldc(self, row):
        return row["ldc"]

    def map_is_reversible(self, row):
        return row["reversible"]

    def map_controlled_bus(self, row):
        bus_name = row["bus"]
        return self.system.get_component(component_type=DistributionBus, name=bus_name)

    def map_controlled_phase(self, row):
        return phase_mapper[row["phase"]]

    def map_delay(self, row):
        return Time(float(row["initial_delay"]), "second")

    def map_vsetpoint(self, row):
        return Voltage(float(row["target"]), "volt")

    def min_v_limit(self, row):
        return Voltage(float(row["min_voltage"]), "volt")

    def max_v_limit(self, row):
        return Voltage(float(row["max_voltage"]), "volt")

    def map_pt_ratio(self, row):
        return float(row["pt_ratio"])

    def map_ldc_R(self, row):
        return float(row["line_drop_r"])

    def map_ldc_X(self, row):
        return float(row["line_drop_x"])

    def map_ct_primary(self, row):
        return float(row["ct_rating"])

    def map_max_step(self, row):
        return 5

    def map_bandwidth(self, row):
        return Voltage(float(row["deadband"]), "volt")



from gdm.distribution.components import MatrixImpedanceSwitch, DistributionBus
from gdm.quantities import Distance
from gdm.distribution.equipment import (
    MatrixImpedanceSwitchEquipment,
    MatrixImpedanceBranchEquipment,
)

from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper


class MatrixImpedanceSwitchMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        self.bus_2 = self.system.get_component(DistributionBus, row["bus_2"])
        self.n_phases = len(self.bus_2.phases)

        return MatrixImpedanceSwitch(
            name=self.map_name(row),
            buses=self.map_buses(row),
            length=Distance(1, "m"),
            phases=self.map_phases(row),
            equipment=self.map_equipment(row),
            is_closed=self.map_is_closed(row),
        )

    def map_is_closed(self, row):
        state = True if row["is_open"] == "false" else False
        return [state] * 3

    def map_name(self, row):
        return row["switch_name"]

    def map_buses(self, row):
        bus_1_name = row["bus_1"]
        bus_1 = self.system.get_component(DistributionBus, bus_1_name)
        bus_2_name = row["bus_2"]
        bus_2 = self.system.get_component(DistributionBus, bus_2_name)
        return [bus_1, bus_2]

    def map_length(self, row):
        length = float(row["length"])
        return Distance(length, "m")

    def map_phases(self, row):
        return self.bus_2.phases

    def map_equipment(self, row):
        equipments: list[MatrixImpedanceBranchEquipment] = list(
            self.system.get_components(
                MatrixImpedanceBranchEquipment,
                filter_func=lambda x: len(x.r_matrix) == self.n_phases,
            )
        )
        if len(equipments):
            model_dict = equipments[0].model_dump()
            equipment = MatrixImpedanceSwitchEquipment(**model_dict)
            return equipment
        else:
            raise Exception(
                "No Matrix Impedance Branch Equipment found with {} phases".format(self.n_phases)
            )



from gdm.quantities import ActivePower, ReactivePower
from gdm.distribution.enums import ConnectionType

from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class DistributionLoadMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "Load_kWkvar"
    altdss_composition_name = "Load"
    opendss_file = OpenDSSFileTypes.LOADS_FILE.value

    def map_in_service(self):
        self.opendss_dict["enabled"] = self.model.in_service

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name
        # TODO: Want to set the Yearly attribute here, but need to access the system. Is that possible?

    def map_bus(self):
        num_phases = len(self.model.phases)
        self.opendss_dict["Bus1"] = self.model.bus.name
        for phase in self.model.phases:
            self.opendss_dict["Bus1"] += self.phase_map[phase]
        # TODO: Should we include the phases its connected to here?
        nom_voltage = self.model.bus.rated_voltage.to("kV").magnitude
        self.opendss_dict["kV"] = nom_voltage if num_phases == 1 else nom_voltage * 1.732

    def map_phases(self):
        if (
            len(self.model.phases) == 2
            and self.model.equipment.connection_type == ConnectionType.DELTA
        ):
            self.opendss_dict["Phases"] = 1
        else:
            self.opendss_dict["Phases"] = len(self.model.phases)
        # TODO: Do we need to remove neutrals?

    def map_equipment(self):
        # TODO: We're not building equipment for the Loads. This means that there's no guarantee that we're addressing all of the attributes in the equipment in a structured way like we are for the component.
        equipment = self.model.equipment
        connection = self.connection_map[equipment.connection_type]
        self.opendss_dict["Conn"] = connection
        self.opendss_dict["Model"] = 8  # Using ZIP model

        # Solve for the co-efficients of Z, I and P terms summed accross all phases A, B and C:
        # P*a_p = P_{0A} * a_{Ap} + P_{0B} * a_{Bp} + P_{0C} * a_{Cp} = z_real
        # P*b_p = P_{0A} * b_{Ap} + P_{0B} * b_{Bp} + P_{0C} * b_{Cp} = i_real
        # P*c_p = P_{0A} * c_{Ap} + P_{0B} * c_{Bp} + P_{0C} * c_{Cp} = p_real
        # a_p + b_p + c_p = 1
        #
        # Gives:
        # P = z_real + i_real + p_real
        # a_p = z_real/ (z_real + i_real + p_real)
        # b_p = i_real/ (z_real + i_real + p_real)
        # c_p = p_real/ (z_real + i_real + p_real)
        #
        # Similar logic for reactive power.

        z_real = ActivePower(0, "kilowatt")
        i_real = ActivePower(0, "kilowatt")
        p_real = ActivePower(0, "kilowatt")
        z_imag = ReactivePower(0, "kilovar")
        i_imag = ReactivePower(0, "kilovar")
        p_imag = ReactivePower(0, "kilovar")
        for phase_load in equipment.phase_loads:
            z_real += phase_load.real_power * phase_load.z_real
            i_real += phase_load.real_power * phase_load.i_real
            p_real += phase_load.real_power * phase_load.p_real
            z_imag += phase_load.reactive_power * phase_load.z_imag
            i_imag += phase_load.reactive_power * phase_load.i_imag
            p_imag += phase_load.reactive_power * phase_load.p_imag
        z_real = z_real.to("kW")
        i_real = i_real.to("kW")
        p_real = p_real.to("kW")
        z_imag = z_imag.to("kvar")
        i_imag = i_imag.to("kvar")
        p_imag = p_imag.to("kvar")
        p_total = z_real + i_real + p_real
        q_total = z_imag + i_imag + p_imag

        if p_total.magnitude != 0:
            a_p = z_real.magnitude / p_total.magnitude
            b_p = i_real.magnitude / p_total.magnitude
            c_p = p_real.magnitude / p_total.magnitude
        else:
            # Degenerate solution - pick a_p = 1 arbitrarily in P=0 case
            a_p = 1.0
            b_p = 0.0
            c_p = 0.0

        if q_total.magnitude != 0:
            a_q = z_imag.magnitude / q_total.magnitude
            b_q = i_imag.magnitude / q_total.magnitude
            c_q = p_imag.magnitude / q_total.magnitude
        else:
            # Degenerate solution - pick a_q = 1 arbitrarily in Q=0 case
            a_q = 1.0
            b_q = 0.0
            c_q = 0.0

        self.opendss_dict["kW"] = p_total.magnitude
        self.opendss_dict["kvar"] = q_total.magnitude
        # Cut-off voltage set to 0
        self.opendss_dict["ZIPV"] = [a_p, b_p, c_p, a_q, b_q, c_q, 0]



from gdm.distribution.components import DistributionBus, DistributionRegulator
from gdm.distribution.controllers import RegulatorController
from gdm.distribution.enums import Phase

from ditto.readers.cim_iec_61968_13.equipment.distribution_transformer_equipment import (
    DistributionTransformerEquipmentMapper,
)

from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper
from ditto.readers.cim_iec_61968_13.common import phase_mapper


class DistributionRegulatorMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        return DistributionRegulator(
            name=self.map_name(row),
            buses=self.map_bus(row),
            equipment=self.map_equipment(row),
            winding_phases=self.map_winding_phases(row),
            controllers=self.map_controllers(row),
        )

    def map_name(self, row):
        return row["xfmr"]

    def map_winding_phases(self, row):
        if "wdg_1_phase" in row:
            phase_1 = [phase_mapper[phs] for phs in row["wdg_1_phase"].replace("N", "")]
        else:
            phase_1 = [Phase.A, Phase.B, Phase.C]

        if "wdg_2_phase" in row:
            phase_2 = [phase_mapper[phs] for phs in row["wdg_2_phase"].replace("N", "")]
        else:
            phase_2 = [Phase.A, Phase.B, Phase.C]

        return [phase_1, phase_2]

    def map_bus(self, row):
        bus_1_name = row["bus_1"]
        bus_2_name = row["bus_2"]
        bus_1 = self.system.get_component(DistributionBus, bus_1_name)
        bus_2 = self.system.get_component(DistributionBus, bus_2_name)
        return [bus_1, bus_2]

    def map_equipment(self, row):
        xfmr_equip_mapper = DistributionTransformerEquipmentMapper(self.system)
        xfmr_equip = xfmr_equip_mapper.parse(row)
        return xfmr_equip

    def map_controllers(self, row):
        reg_controllers = self.system.get_component(RegulatorController, row["xfmr"])
        return [reg_controllers]



from enum import Enum


class OpenDSSFileTypes(str, Enum):
    MASTER_FILE = "Master.dss"
    COORDINATE_FILE = "BusCoords.dss"
    WIRES_FILE = "WireData.dss"
    LINE_GEOMETRIES_FILE = "LineGeometry.dss"
    LINECODES_FILE = "LineCodes.dss"
    SWITCH_CODES_FILE = "SwitchCodes.dss"
    FUSE_CODES_FILE = "FuseCodes.dss"
    TRANSFORMERS_FILE = "Transformers.dss"
    CAPACITORS_FILE = "Capacitors.dss"
    LINES_FILE = "Lines.dss"
    LOADS_FILE = "Loads.dss"
    SWITCH_FILE = "Switches.dss"
    FUSE_FILE = "Fuses.dss"
    REGULATOR_CONTROLLERS_FILE = "RegControllers.dss"



from gdm.quantities import ReactivePower, Resistance, Reactance
from gdm.distribution.equipment.phase_capacitor_equipment import PhaseCapacitorEquipment
from gdm.distribution.equipment.capacitor_equipment import CapacitorEquipment
from gdm.distribution.enums import ConnectionType, VoltageTypes
from gdm.quantities import Voltage

from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper


class CapacitorEquipmentMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        phases = row["phase"]
        if phases is None:
            self.phases = ["A", "B", "C"]
        else:
            self.phases = phases.split(",")

        return CapacitorEquipment(
            name=self.map_name(row),
            phase_capacitors=self.map_phase_capacitors(row),
            connection_type=self.map_connection_type(row),
            rated_voltage=self.map_rated_voltage(row),
            voltage_type=self.map_voltage_type(),
        )

    def map_name(self, row):
        return row["capacitor"] + "_equipment"

    def map_voltage_type(self):
        n_phases = len(self.phases)
        return VoltageTypes.LINE_TO_LINE if n_phases == 3 else VoltageTypes.LINE_TO_LINE

    def map_rated_voltage(self, row):
        return Voltage(float(row["rated_voltage"]), "volt")

    def map_phase_capacitors(self, row):
        phase_loads = []
        n_phases = len(self.phases)
        voltage = (
            float(row["rated_voltage"]) if n_phases == 3 else float(row["rated_voltage"]) / 1.732
        )
        b1 = float(row["b1"])
        var = voltage**2 * b1
        var_per_phase = var / n_phases
        for phase in self.phases:
            if b1 > 0:
                mapper = PhaseCapacitorEquipmentMapper(self.system)
                phase_load = mapper.parse(row, var_per_phase, phase)
                phase_loads.append(phase_load)
        return phase_loads

    def map_connection_type(self, row):
        return ConnectionType.DELTA if row["conn"] == "D" else ConnectionType.STAR


class PhaseCapacitorEquipmentMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row, var_per_phase, phase):
        return PhaseCapacitorEquipment(
            name=self.map_name(row, phase),
            resistance=self.map_resistance(),
            reactance=self.map_reactance(),
            rated_reactive_power=self.map_rated_reactive_power(var_per_phase),
            num_banks_on=self.map_num_banks_on(row),
            num_banks=self.map_num_banks(row),
        )

    def map_name(self, row, phase):
        return row["capacitor"] + "_phase_capacitor_equipment_" + phase

    # Resistance and Reactance not included for capacitors
    def map_resistance(self):
        return Resistance(0, "ohm")

    # Resistance and Reactance not included for capacitors
    def map_reactance(self):
        return Reactance(0, "ohm")

    # TODO: This doesn't make sense. We should have fixed and switched values
    def map_rated_reactive_power(self, var_per_phase):
        return ReactivePower(var_per_phase, "var")

    # TODO: This doesn't make sense. This should indicate if the bank is switched
    def map_num_banks_on(self, row):
        return row["steps"]

    # TODO: This doesn't make sense. This should indicate how many banks are switched
    def map_num_banks(self, row):
        return row["steps"]



from gdm.distribution.enums import VoltageTypes

from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class DistributionVoltageSourceMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "Vsource_Z0Z1Z2"
    altdss_composition_name = "Vsource"
    opendss_file = OpenDSSFileTypes.MASTER_FILE.value

    def map_in_service(self):
        self.opendss_dict["enabled"] = self.model.in_service

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_bus(self):
        self.opendss_dict["Bus1"] = self.model.bus.name
        for phase in self.model.phases:
            self.opendss_dict["Bus1"] += self.phase_map[phase]

    def map_phases(self):
        # Handled in the map_bus function
        self.opendss_dict["Phases"] = len(self.model.phases)
        return

    def map_equipment(self):
        r1 = 0
        x1 = 0
        r0 = 0
        x0 = 0
        voltage = self.model.equipment.sources[0].voltage
        angle = self.model.equipment.sources[0].angle
        num_phases = len(self.model.phases)
        rated_voltage = self.model.bus.rated_voltage

        for phase_source in self.model.equipment.sources:
            r1 += phase_source.r1
            r0 += phase_source.r0
            x1 += phase_source.x1
            x0 += phase_source.x0
            break

        r1 = r1.to("ohm")
        r0 = r0.to("ohm")
        x1 = x1.to("ohm")
        x0 = x0.to("ohm")
        voltage = voltage.to("kilovolt")
        rated_voltage = rated_voltage.to("kilovolt")
        angle = angle.to("degree")

        if self.model.equipment.sources[0].voltage_type == VoltageTypes.LINE_TO_GROUND:
            if num_phases == 1:
                v_mag = voltage.magnitude
            else:
                v_mag = voltage.magnitude * 1.732
        else:
            if num_phases == 1:
                v_mag = voltage.magnitude / 1.732
            else:
                v_mag = voltage.magnitude

        v_nom = rated_voltage.magnitude if num_phases == 1 else rated_voltage.magnitude * 1.732
        self.opendss_dict["Angle"] = angle.magnitude
        self.opendss_dict["pu"] = v_mag / v_nom
        self.opendss_dict["BasekV"] = v_nom
        self.opendss_dict["Z0"] = complex(r0.magnitude, x0.magnitude)
        self.opendss_dict["Z1"] = complex(r1.magnitude, x1.magnitude)



from ditto.writers.opendss.components.distribution_bus import DistributionBusMapper
from ditto.writers.opendss.components.distribution_branch import DistributionBranchMapper
from ditto.writers.opendss.components.sequence_impedance_branch import (
    SequenceImpedanceBranchMapper,
)
from ditto.writers.opendss.components.matrix_impedance_branch import MatrixImpedanceBranchMapper
from ditto.writers.opendss.components.geometry_branch import GeometryBranchMapper
from ditto.writers.opendss.equipment.sequence_impedance_branch_equipment import (
    SequenceImpedanceBranchEquipmentMapper,
)
from ditto.writers.opendss.equipment.matrix_impedance_branch_equipment import (
    MatrixImpedanceBranchEquipmentMapper,
)
from ditto.writers.opendss.equipment.geometry_branch_equipment import GeometryBranchEquipmentMapper
from ditto.writers.opendss.equipment.bare_conductor_equipment import BareConductorEquipmentMapper
from ditto.writers.opendss.components.distribution_capacitor import DistributionCapacitorMapper
from ditto.writers.opendss.components.distribution_load import DistributionLoadMapper
from ditto.writers.opendss.components.distribution_transformer import DistributionTransformerMapper
from ditto.writers.opendss.equipment.distribution_transformer_equipment import (
    DistributionTransformerEquipmentMapper,
)
from ditto.writers.opendss.components.distribution_vsource import DistributionVoltageSourceMapper
from ditto.writers.opendss.equipment.matrix_impedance_switch_equipment import (
    MatrixImpedanceSwitchEquipmentMapper,
)
from ditto.writers.opendss.components.matrix_impedance_switch import MatrixImpedanceSwitchMapper
from ditto.writers.opendss.equipment.matrix_impedance_fuse_equipment import (
    MatrixImpedanceFuseEquipmentMapper,
)
from ditto.writers.opendss.components.matrix_impedance_fuse import MatrixImpedanceFuseMapper
from ditto.writers.opendss.components.distribution_regulator import DistributionRegulatorMapper
from ditto.writers.opendss.controllers.distribution_regulator_controller import (
    RegulatorControllerMapper,
)



from collections import Counter
from ast import literal_eval
from typing import Any
from uuid import uuid4
from enum import Enum

from gdm.quantities import ApparentPower, Voltage
from gdm.distribution.common import SequencePair
from gdm.distribution.components import (
    DistributionTransformer,
    DistributionBus,
)
from gdm.distribution.equipment import (
    DistributionTransformerEquipment,
    WindingEquipment,
)
from gdm.distribution.enums import (
    ConnectionType,
    VoltageTypes,
    Phase,
)
from infrasys.system import System
import opendssdirect as odd
from loguru import logger

from ditto.readers.opendss.common import PHASE_MAPPER, get_equipment_from_catalog

SEQUENCE_PAIRS = [SequencePair(1, 2), SequencePair(1, 3), SequencePair(2, 3)]


class XfmrModelTypes(str, Enum):
    TRANSFORMERS = "Transformer"
    XFMRCODE = "XfmrCode"


def _build_xfmr_equipment(
    model_type: str,
    distribution_transformer_equipment_catalog: dict[int, DistributionTransformerEquipment],
    winding_equipment_catalog: dict[int, WindingEquipment],
) -> tuple[DistributionTransformerEquipment, list[DistributionBus], list[Phase]]:
    """Helper function to build a DistributionTransformerEquipment instance

    Args:
        model_type (str): Opendss model type e.g. Transformer, XfmrCode
        distribution_transformer_equipment_catalog (dict[int, DistributionTransformerEquipment]): mapping of model hash to DistributionTransformerEquipment instance
        winding_equipment_catalog (dict[int, WindingEquipment]): mapping of model hash to WindingEquipment instance

    Returns:
        DistributionTransformerEquipment: instance of DistributionTransformerEquipment
        list[DistributionBus]: List of DistributionBus
        list[Phase]: List of Phase
    """

    model_name = odd.Element.Name().lower().split(".")[1]
    if model_type == XfmrModelTypes.XFMRCODE.value:
        equipment_uuid = model_name
    else:
        equipment_uuid = str(uuid4())

    def query(property: str, dtype: type):
        command = f"? {model_type}.{model_name}.{property}"
        odd.Text.Command(command)
        result = odd.Text.Result()

        if result is None:
            if dtype in [float, int]:
                return 0
            elif dtype is str:
                return ""
            else:
                return None
        else:
            if dtype is list:
                return literal_eval(result)
            else:
                return dtype(result)

    def set_ppty(property: str, value: Any):
        return odd.Command(f"{model_type}.{model_name}.{property}={value}")

    all_reactances = [
        query("xhl", float),
        query("xht", float),
        query("xlt", float),
    ]

    number_windings = query("windings", int)
    wdg_nom_voltages = []
    windings = []

    for wdg_index in range(number_windings):
        set_ppty("Wdg", wdg_index + 1)
        num_phase = query("phases", int)
        if query("conn", str).lower() == "delta":
            rated_voltage = query("kv", float) / 1.732
        else:
            rated_voltage = query("kv", float) / 1.732 if num_phase == 3 else query("kv", float)
        wdg_nom_voltages.append(rated_voltage)
        min_tap_pu = query("mintap", float)
        max_tap_pu = query("maxtap", float)
        num_taps = query("numtaps", int)
        taps = query("taps", list)
        tap = [taps[wdg_index]] * num_phase
        winding = WindingEquipment.model_construct(
            rated_power=ApparentPower(query("kva", float), "kilova"),
            num_phases=num_phase,
            connection_type=ConnectionType.DELTA
            if query("conn", str).lower() == "delta"
            else ConnectionType.STAR,
            rated_voltage=Voltage(rated_voltage, "kilovolt"),
            resistance=query("%r", float),
            is_grounded=False,  # TODO: Should be moved to the transformer model. Only known once the transformer is installed
            voltage_type=VoltageTypes.LINE_TO_GROUND,
            tap_positions=tap,
            total_taps=num_taps,
            min_tap_pu=min_tap_pu,
            max_tap_pu=max_tap_pu,
        )
        winding = get_equipment_from_catalog(winding, winding_equipment_catalog)
        windings.append(winding)

    coupling_sequences = SEQUENCE_PAIRS[:1] if number_windings == 2 else SEQUENCE_PAIRS
    reactances = all_reactances[:1] if number_windings == 2 else all_reactances

    dist_transformer = DistributionTransformerEquipment.model_construct(
        name=equipment_uuid,
        pct_no_load_loss=query(r"%noloadloss", float),
        pct_full_load_loss=query(r"%loadloss", float),
        windings=windings,
        coupling_sequences=coupling_sequences,
        winding_reactances=reactances,
        is_center_tapped=_is_center_tapped(wdg_nom_voltages),
    )
    dist_transformer = get_equipment_from_catalog(
        dist_transformer, distribution_transformer_equipment_catalog
    )
    return dist_transformer


def get_transformer_equipments(system: System) -> list[DistributionTransformerEquipment]:
    """Function to return list of all DistributionTransformerEquipment in Opendss model.

    Args:
        system (System): Instance of infrasys System

    Returns:
        list[DistributionTransformerEquipment]: List of DistributionTransformerEquipment objects
    """
    logger.debug("parsing transformer equipment...")
    distribution_transformer_equipment_catalog = {}
    winding_equipment_catalog = {}
    odd_model_types = [v.value for v in XfmrModelTypes]
    for odd_model_type in odd_model_types:
        odd.Circuit.SetActiveClass(odd_model_type)
        flag = odd.ActiveClass.First()
        while flag > 0:
            _build_xfmr_equipment(
                odd_model_type,
                distribution_transformer_equipment_catalog,
                winding_equipment_catalog,
            )
            flag = odd.ActiveClass.Next()
    return distribution_transformer_equipment_catalog, winding_equipment_catalog


def get_transformers(
    system: System,
    distribution_transformer_equipment_catalog: dict[int, DistributionTransformerEquipment],
    winding_equipment_catalog: dict[int, WindingEquipment],
) -> list[DistributionTransformer]:
    """Method returns a list of DistributionTransformer objects

    Args:
        system (System):  Instance of System
        distribution_transformer_equipment_catalog (dict[int, DistributionTransformerEquipment]): mapping of model hash to DistributionTransformerEquipment instance
        winding_equipment_catalog (dict[int, WindingEquipment]): mapping of model hash to WindingEquipment instance

     Returns:
        list[DistributionTransformer]: list of distribution transformers
    """

    logger.debug("parsing transformer components...")

    transformers = []
    flag = odd.Transformers.First()
    while flag > 0:
        logger.debug(f"building transformer {odd.Transformers.Name()}")
        bus_names = odd.CktElement.BusNames()
        buses = []
        phases = []
        for bus_name in bus_names:
            bus_name_clean = bus_name.split(".")[0]
            bus_phases = bus_name.split(".")[1:] if "." in bus_name else ["1", "2", "3"]
            bus_phases = [PHASE_MAPPER[ph] for ph in bus_phases]
            phases.append([Phase(phase) for phase in bus_phases])
            buses.append(system.get_component(DistributionBus, bus_name_clean))
        xfmr_equipment = _build_xfmr_equipment(
            XfmrModelTypes.TRANSFORMERS.value,
            distribution_transformer_equipment_catalog,
            winding_equipment_catalog,
        )
        transformer = DistributionTransformer.model_construct(
            name=odd.Transformers.Name().lower(),
            buses=buses,
            winding_phases=phases,
            equipment=xfmr_equipment,
        )
        transformers.append(transformer)
        flag = odd.Transformers.Next()

    return transformers


def _is_center_tapped(wdg_nom_voltages: list[float]) -> bool:
    """The flag is true if the transformer is center tapped

    Args:
        wdg_nom_voltages (list[float]): list of nomimal voltage for each bus

    Returns:
        bool: True if the transfomer equpment is split phase else False
    """

    is_split_phase = False
    if len(wdg_nom_voltages) == 3:
        max_secondary_voltage_kv = 0.6
        counts = Counter(wdg_nom_voltages)
        for nom_voltage in counts:
            if nom_voltage < max_secondary_voltage_kv and counts[nom_voltage] == 2:
                is_split_phase = True
    return is_split_phase



from uuid import uuid4

from gdm.quantities import ActivePower, ReactivePower
from gdm.distribution.enums import ConnectionType
from gdm.distribution.components import (
    DistributionBus,
    DistributionLoad,
)
from gdm.distribution.equipment import (
    PhaseLoadEquipment,
    LoadEquipment,
)
from infrasys.system import System
import opendssdirect as odd
from loguru import logger

from ditto.readers.opendss.components.loadshapes import build_profiles, ObjectsWithProfile
from ditto.readers.opendss.common import PHASE_MAPPER, LoadTypes


def _build_load_equipment() -> tuple[LoadEquipment, list[str], str, list[str]]:
    """Helper function to build a LoadEquipment instance

    Returns:
        LoadEquipment: instance of LoadEquipment
        list[str]: List of buses
        list[str]: List of phases
    """

    equipment_uuid = uuid4()
    buses = odd.CktElement.BusNames()
    num_phase = odd.CktElement.NumPhases()
    kvar_ = odd.Loads.kvar()
    kw_ = odd.Loads.kW()
    zip_params = odd.Loads.ZipV()
    model = odd.Loads.Model()
    nodes = buses[0].split(".")[1:] if num_phase != 3 else ["1", "2", "3"]
    ph_loads = []
    zip_param_keys = ["z_real", "i_real", "p_real", "z_imag", "i_imag", "p_imag"]
    for el in nodes:
        kw_per_phase = kw_ / len(nodes)
        kvar_per_phase = kvar_ / len(nodes)
        zip_params_dict = {param: 0 for param in zip_param_keys}
        if model == LoadTypes.CONST_POWER:
            zip_params_dict.update({"p_imag": 1, "p_real": 1})
        elif model == LoadTypes.CONST_CURRENT:
            zip_params_dict.update({"i_imag": 1, "i_real": 1})
        elif model == LoadTypes.CONST_IMPEDANCE:
            zip_params_dict.update({"z_imag": 1, "z_real": 1})
        elif model == LoadTypes.ZIP:
            zip_params_dict = dict(zip(zip_param_keys, zip_params))
        elif model == LoadTypes.CONST_P__QUARDRATIC_Q:
            zip_params_dict.update({"p_real": 1, "z_imag": 1})
        elif model == LoadTypes.LINEAR_P__QUARDRATIC_Q:
            zip_params_dict.update({"i_real": 1, "z_imag": 1})
        else:
            msg = f"Invalid load model type {model} passed. valid options are {LoadTypes}"
            raise ValueError(msg)
        load = PhaseLoadEquipment.model_construct(
            name=f"{equipment_uuid}_{el}",
            real_power=ActivePower(kw_per_phase, "kilowatt"),
            reactive_power=ReactivePower(kvar_per_phase, "kilovar"),
            **zip_params_dict,
        )
        ph_loads.append(load)
    load_equipment = LoadEquipment.model_construct(
        name=str(uuid4()),
        phase_loads=ph_loads,
        connection_type=ConnectionType.DELTA if odd.Loads.IsDelta() else ConnectionType.STAR,
    )
    return load_equipment, buses, nodes


# def get_load_equipments(odd:Opendssdirect) -> list[LoadEquipment]:
#     """Function to return list of all LoadEquipment in Opendss model.

#     Args:
#         odd (Opendssdirect): Instance of Opendss simulator

#     Returns:
#         list[LoadEquipment]: List of LoadEquipment objects
#     """


def get_loads(system: System) -> list[DistributionLoad]:
    """Function to return list of DistributionLoad in Opendss model.

    Args:
        system (System): Instance of System

    Returns:
        List[DistributionLoad]: List of DistributionLoad objects
    """

    logger.debug("parsing load components...")

    profile_catalog = {}
    loads = []
    flag = odd.Loads.First()
    while flag > 0:
        load_name = odd.Loads.Name().lower()
        LoadEquipment, buses, nodes = _build_load_equipment()
        bus1 = buses[0].split(".")[0]
        profile_names = [odd.Loads.Daily(), odd.Loads.Yearly(), odd.Loads.Duty()]
        profiles = build_profiles(profile_names, ObjectsWithProfile.LOAD, profile_catalog)
        distribution_load = DistributionLoad.model_construct(
            name=load_name,
            bus=system.get_component(DistributionBus, bus1),
            phases=[PHASE_MAPPER[el] for el in nodes],
            equipment=LoadEquipment,
        )
        for profile_name in profile_names:
            if profile_name in profiles:
                for profile_type, ts_profile in profiles[profile_name].items():
                    system.add_time_series(
                        ts_profile["data"],
                        distribution_load,
                        profile_type=profile_type,
                        profile_name=profile_name,
                        use_actual=ts_profile["use_actual"],
                    )
                    logger.debug(
                        f"Adding timeseries profile '{profile_name} / {profile_type}' to load '{load_name}'"
                    )
        loads.append(distribution_load)
        flag = odd.Loads.Next()
    return loads



from uuid import uuid4

from gdm.distribution.enums import VoltageTypes, ConnectionType
from gdm.distribution.components import DistributionCapacitor, DistributionBus
from gdm.distribution.equipment import PhaseCapacitorEquipment, CapacitorEquipment
from gdm.quantities import (
    ReactivePower,
    Resistance,
    Reactance,
    Voltage,
)
from infrasys.system import System
import opendssdirect as odd
from loguru import logger

from ditto.readers.opendss.common import PHASE_MAPPER, get_equipment_from_catalog


def _build_capacitor_source_equipment(
    phase_capacitor_equipment_catalog: dict[int, PhaseCapacitorEquipment],
    capacitor_equipment_catalog: dict[int, CapacitorEquipment],
) -> tuple[CapacitorEquipment, list[str], list[str]]:
    """Helper function to build a CapacitorEquipment instance

    Args:
        phase_capacitor_equipment_catalog (dict[int, PhaseCapacitorEquipment]): mapping of model hash to PhaseCapacitorEquipment instance
        capacitor_equipment_catalog (dict[int, CapacitorEquipment]):  mapping of model hash to CapacitorEquipment instance

    Returns:
        CapacitorEquipment: Instance of CapacitorEquipment
        list[str]: List of buses
        list[str]: List of phases
    """

    equipment_uuid = uuid4()
    buses = odd.CktElement.BusNames()
    num_phase = odd.CktElement.NumPhases()
    kvar_ = odd.Capacitors.kvar()

    ph_caps = []
    nodes = buses[0].split(".")[1:] if num_phase != 3 else ["1", "2", "3"]
    live_nodes = [node for node in nodes if node != "0"]

    voltage_type = (
        VoltageTypes.LINE_TO_GROUND
        if num_phase == 1 and len(live_nodes)
        else VoltageTypes.LINE_TO_LINE
    )

    for el in nodes:
        phase_capacitor = PhaseCapacitorEquipment.model_construct(
            name=f"{equipment_uuid}_{el}",
            rated_reactive_power=ReactivePower(kvar_ / len(nodes), "kilovar"),
            num_banks=odd.Capacitors.NumSteps(),
            num_banks_on=sum(odd.Capacitors.States()),
            resistance=Resistance(0, "ohm"),
            reactance=Reactance(0, "ohm"),
        )
        phase_capacitor = get_equipment_from_catalog(
            phase_capacitor, phase_capacitor_equipment_catalog
        )
        ph_caps.append(phase_capacitor)

    capacitor_equipment = CapacitorEquipment.model_construct(
        name=str(equipment_uuid),
        phase_capacitors=ph_caps,
        connection_type=ConnectionType.DELTA if odd.Capacitors.IsDelta() else ConnectionType.STAR,
        rated_voltage=Voltage(odd.Capacitors.kV(), "kilovolt"),
        voltage_type=voltage_type,
    )

    capacitor_equipment = get_equipment_from_catalog(
        capacitor_equipment, capacitor_equipment_catalog
    )

    return capacitor_equipment, buses, nodes


def get_capacitors(system: System) -> list[DistributionCapacitor]:
    """Function to return list of all capacitors in Opendss model.

    Args:
        system (System): Instance of System

    Returns:
        List[DistributionCapacitor]: List of DistributionCapacitor objects
    """

    logger.debug("parsing capacitor components...")
    phase_capacitor_equipment_catalog = {}
    capacitor_equipment_catalog = {}
    capacitors = []
    flag = odd.Capacitors.First()
    while flag > 0:
        equipment, buses, nodes = _build_capacitor_source_equipment(
            phase_capacitor_equipment_catalog, capacitor_equipment_catalog
        )
        bus1 = buses[0].split(".")[0]
        capacitors.append(
            DistributionCapacitor.model_construct(
                name=odd.Capacitors.Name().lower(),
                bus=system.get_component(DistributionBus, bus1),
                phases=[PHASE_MAPPER[el] for el in nodes],
                controllers=[],
                equipment=equipment,
            )
        )
        flag = odd.Capacitors.Next()
    return capacitors



from ditto.writers.opendss.equipment.matrix_impedance_branch_equipment import (
    MatrixImpedanceBranchEquipmentMapper,
)
from ditto.enumerations import OpenDSSFileTypes


class MatrixImpedanceSwitchEquipmentMapper(MatrixImpedanceBranchEquipmentMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "LineCode_ZMatrixCMatrix"
    altdss_composition_name = "LineCode"
    opendss_file = OpenDSSFileTypes.SWITCH_CODES_FILE.value

    def map_controller(self):
        # Not mapped in OpenDSS
        pass



from gdm.distribution.equipment.distribution_transformer_equipment import (
    DistributionTransformerEquipment,
)
from gdm.distribution.common import SequencePair

from ditto.readers.cim_iec_61968_13.equipment.winding_equipment import WindingEquipmentMapper
from ditto.readers.cim_iec_61968_13.cim_mapper import CimMapper


class DistributionTransformerEquipmentMapper(CimMapper):
    def __init__(self, system):
        super().__init__(system)

    def parse(self, row):
        self.s = float(row["wdg_1_apparent_power"])
        self.v_h = float(row["wdg_1_rated_voltage"])
        self.v_l = float(row["wdg_2_rated_voltage"])
        self.r_h = float(row["wdg_1_per_resistance"])
        self.r_l = float(row["wdg_2_per_resistance"])

        self.per_r_1 = self.r_h / (self.v_h**2 / self.s) * 100
        self.per_r_2 = self.r_l / (self.v_l**2 / self.s) * 100

        return DistributionTransformerEquipment(
            name=self.map_name(row),
            pct_no_load_loss=self.map_pct_no_load_loss(row),
            pct_full_load_loss=self.map_pct_full_load_loss(row),
            windings=self.map_windings(row),
            coupling_sequences=self.map_coupling_sequences(row),
            winding_reactances=self.map_winding_reactances(row),
            is_center_tapped=self.map_is_center_tapped(row),
        )

    def map_name(self, row):
        return row["xfmr"] + "_equipment"

    def map_pct_no_load_loss(self, row):
        return 0

    def map_pct_full_load_loss(self, row):
        return self.per_r_1 + self.per_r_2

    def map_windings(self, row):
        mapper = WindingEquipmentMapper(self.system)
        wingings = mapper.parse(row)
        return wingings

    def map_coupling_sequences(self, row):
        return [SequencePair(1, 2)]

    def map_winding_reactances(self, row):
        if "x0" in row and "x1" in row:
            x0 = float(row["x0"])
            x1 = float(row["x1"])
            x_hl = 1 / 3 * (2 * x1 + x0)
            per_x = x_hl / (self.v_h**2 / self.s) * 100

            return [per_x]

        else:
            return [1]

    def map_is_center_tapped(self, row):
        return False



from gdm.distribution.enums import Phase

from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class DistributionBranchMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "Line_Common"
    altdss_composition_name = "Line"
    opendss_file = OpenDSSFileTypes.LINES_FILE.value

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_buses(self):
        self.opendss_dict["Bus1"] = self.model.buses[0].name
        self.opendss_dict["Bus2"] = self.model.buses[1].name
        for phase in self.model.phases:
            if phase != Phase.N:
                self.opendss_dict["Bus1"] += self.phase_map[phase]
                self.opendss_dict["Bus2"] += self.phase_map[phase]

    def map_length(self):
        self.opendss_dict["Length"] = self.model.length.magnitude
        model_unit = str(self.model.length.units)
        if model_unit not in self.length_units_map:
            raise ValueError(f"{model_unit} not mapped for OpenDSS")
        self.opendss_dict["Units"] = self.length_units_map[model_unit]

    def map_phases(self):
        # Redundant information - included in buses
        # TODO: remove from GDM?

        live_phases = [phase for phase in self.model.phases if phase != Phase.N]
        self.opendss_dict["Phases"] = len(live_phases)
        pass



from gdm.distribution.equipment import BareConductorEquipment, ConcentricCableEquipment

from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class GeometryBranchEquipmentMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "LineGeometry_xh"
    altdss_composition_name = "LineGeometry"
    opendss_file = OpenDSSFileTypes.LINECODES_FILE.value

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_common(self):
        units = []
        assert len(self.model.horizontal_positions) == len(self.model.vertical_positions)

        for i in range(len(self.model.horizontal_positions)):
            horizontal_position = self.model.horizontal_positions[i]
            vertical_position = self.model.vertical_positions[i]
            x_unit = str(horizontal_position.units)
            h_unit = str(vertical_position.units)
            assert h_unit == x_unit
            if x_unit not in self.length_units_map:
                raise ValueError(f"{x_unit} not mapped for OpenDSS")
            units.append(self.length_units_map[x_unit])
        self.opendss_dict["Units"] = units
        self.opendss_dict["NConds"] = len(self.model.conductors)

    def map_horizontal_positions(self):
        all_x = []
        for horizontal_position in self.model.horizontal_positions:
            all_x.append(horizontal_position.magnitude)
        self.opendss_dict["X"] = all_x

    def map_vertical_positions(self):
        all_h = []
        for vertical_position in self.model.vertical_positions:
            all_h.append(vertical_position.magnitude)
        self.opendss_dict["H"] = all_h

    def map_conductors(self):
        all_conductors = []
        for conductor in self.model.conductors:
            if isinstance(conductor, BareConductorEquipment):
                conductor_type = "wiredata"
            elif isinstance(conductor, ConcentricCableEquipment):
                conductor_type = "cndata"
            #            if isinstance(conductor,TapeShieldCableEqupment):
            #                conductor_type = 'tsdata'
            else:
                raise ValueError(f"Unknown conductor type {conductor}")
            all_conductors.append(f"{conductor_type}.{conductor.name}")
        self.opendss_dict["Conductors"] = all_conductors



from ditto.writers.opendss.opendss_mapper import OpenDSSMapper
from ditto.enumerations import OpenDSSFileTypes


class BareConductorEquipmentMapper(OpenDSSMapper):
    def __init__(self, model):
        super().__init__(model)

    altdss_name = "WireData"
    altdss_composition_name = None
    opendss_file = OpenDSSFileTypes.WIRES_FILE.value

    def map_name(self):
        self.opendss_dict["Name"] = self.model.name

    def map_conductor_diameter(self):
        self.opendss_dict["Radius"] = self.model.conductor_diameter.magnitude / 2
        rad_units = str(self.model.conductor_diameter.units)
        if rad_units not in self.length_units_map:
            raise ValueError(f"{rad_units} not mapped for OpenDSS")
        self.opendss_dict["RadUnits"] = self.length_units_map[rad_units]

    def map_conductor_gmr(self):
        self.opendss_dict["GMRAC"] = self.model.conductor_gmr.magnitude
        gmr_units = str(self.model.conductor_gmr.units)
        if gmr_units not in self.length_units_map:
            raise ValueError(f"{gmr_units} not mapped for OpenDSS")
        self.opendss_dict["GMRUnits"] = self.length_units_map[gmr_units]

    def map_ac_resistance(self):
        resistance = self.model.ac_resistance.to("ohms/km")
        self.opendss_dict["RAC"] = resistance.magnitude
        self.opendss_dict["RUnits"] = "km"

    def map_dc_resistance(self):
        resistance = self.model.dc_resistance.to("ohms/km")
        self.opendss_dict["RDC"] = resistance.magnitude
        self.opendss_dict["RUnits"] = "km"

    def map_ampacity(self):
        ampacity_amps = self.model.ampacity.to("ampere")
        self.opendss_dict["NormAmps"] = ampacity_amps.magnitude

    def map_emergency_ampacity(self):
        ampacity_amps = self.model.ampacity.to("ampere")
        self.opendss_dict["EmergAmps"] = ampacity_amps.magnitude
